import os

import numpy as np
import pandas as pd
import qgis
from qgis.analysis import QgsNativeAlgorithms
from qgis.core import *
from qgis.PyQt.QtCore import *
from joblib import Parallel, delayed
from basinmaker.utilities.utilities import *
import tempfile
from json import load, JSONEncoder
import json
import shutil

def qgis_raster_gdal_warpreproject(processing, Input, TARGET_CRS, Output):
    """Functions reproject raster layer
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    out = processing.run(
        "gdal:warpreproject",
        {
            "INPUT": Input,
            "SOURCE_CRS": None,
            "TARGET_CRS": QgsCoordinateReferenceSystem(TARGET_CRS),
            "RESAMPLING": 0,
            "NODATA": None,
            "TARGET_RESOLUTION": None,
            "OPTIONS": "",
            "DATA_TYPE": 0,
            "TARGET_EXTENT": None,
            "TARGET_EXTENT_CRS": None,
            "MULTITHREADING": False,
            "EXTRA": "",
            "OUTPUT": Output,
        },
    )
    return out

####
def create_geo_jason_file(processing,Input_Polygon_path):


    product_dir = os.path.dirname(Input_Polygon_path)
    Names_in = os.path.basename(Input_Polygon_path).split('_')
    n_charc = len(Names_in)
    version  = Names_in[n_charc - 1][0:4]
    TOLERANCEs = [0.0001,0.0005,0.001,0.005,0.01,0.05]


    head_name_cat = "finalcat_info"
    head_name_riv = "finalcat_info_riv"
    head_name_slake = "sl_connected_lake"
    head_name_nlake = "sl_non_connected_lake"

    Input_file_name = []
    Output_file_name = []                            
    if 'v' in version:
        Input_file_name = [
                           head_name_cat + "_"+version+'.shp',
                           head_name_riv + "_"+version+'.shp',
                           head_name_slake + "_"+version+'.shp',
                           head_name_nlake + "_"+version+'.shp',
                           ]
        Output_file_name = [
                           head_name_cat + "_"+version+'.geojson',
                           head_name_riv + "_"+version+'.geojson',
                           head_name_slake + "_"+version+'.geojson',
                           head_name_nlake + "_"+version+'.geojson',
                           ]
    else:
        Input_file_name = [
                           head_name_cat +'.shp',
                           head_name_riv +'.shp',
                           head_name_slake +'.shp',
                           head_name_nlake +'.shp',
                           ]
        Output_file_name = [
                           head_name_cat +'.geojson',
                           head_name_riv +'.geojson',
                           head_name_slake +'.geojson',
                           head_name_nlake +'.geojson',
                           ]
    created_jason_files = []   
    created_jason_files_lake_riv = [] 
    
    for i  in  range(0,len(Input_file_name)):
        input_path = os.path.join(product_dir,Input_file_name[i])
        output_jason_path = os.path.join(product_dir,Output_file_name[i])
        if not os.path.exists(input_path):
            continue 
        created_jason_files.append(output_jason_path) 
        
        if 'finalcat_info_riv' in Input_file_name[i] or 'connected_lake' in Input_file_name[i]:
            created_jason_files_lake_riv.append(output_jason_path)  
                            
        # reproject to WGS84
        input_wgs_84 = processing.run("native:reprojectlayer", {'INPUT':input_path,
                                      'TARGET_CRS':QgsCoordinateReferenceSystem('EPSG:4326'),
                                       'OUTPUT':"memory:"})['OUTPUT']  

        # input_wgs_84_smth =  processing.run("native:smoothgeometry", {'INPUT':input_wgs_84,
        #                                                         'ITERATIONS':5,'OFFSET':0.5,'MAX_ANGLE':180,
        #                                                         'OUTPUT':"memory:"})['OUTPUT']

        
        if 'finalcat_info' in Input_file_name[i] or "finalcat_info_riv" in Input_file_name[i]:
            input_tojson = processing.run("native:fieldcalculator", {'INPUT':input_wgs_84,
                                                      'FIELD_NAME':'rvhName','FIELD_TYPE':2,'FIELD_LENGTH':20,'FIELD_PRECISION':0,
                                                      'FORMULA':'\'sub\' +to_string(to_int(\"SubId\")) ','OUTPUT':"memory:"})['OUTPUT']
        else:
             input_tojson = input_wgs_84
        
        for TOLERANCE in TOLERANCEs:                               
            input_wgs_84_simplify = processing.run("native:simplifygeometries", {
                                                   'INPUT':input_tojson,
                                                   'METHOD':0,
                                                   'TOLERANCE':TOLERANCE,
                                                   'OUTPUT':"memory:"
                                                   }
                                                   )['OUTPUT']
                                               
            qgis.core.QgsVectorFileWriter.writeAsVectorFormat(input_wgs_84_simplify,output_jason_path, 'utf-8', input_wgs_84_simplify.crs(), 'GeoJson',layerOptions=['COORDINATE_PRECISION=3'])
            
            json_file_size = os.stat(output_jason_path).st_size/1024/1024 #to MB
            if json_file_size <= 100:
                break
    # if len(created_jason_files) > 1 and os.stat(os.path.join(product_dir,Output_file_name[0])).st_size/1024/1024 < 500::
    #     for i in range(0,len(created_jason_files)):
    #         injson = load(open(created_jason_files[i]))
    #         if i == 0:
    #             output_jason = injson
    #         else:
    #             output_jason['features'] += injson['features']
    # 
    #     with open(os.path.join(product_dir,'routing_product.geojson'), 'w', encoding='utf-8') as f:
    #         json.dump(output_jason, f, ensure_ascii=False, indent=4)
    # else:
    #     shutil.copy(created_jason_files[0], os.path.join(product_dir,'routing_product.geojson')) 
        
    if len(created_jason_files_lake_riv) > 1 and os.stat(os.path.join(product_dir,Output_file_name[0])).st_size/1024/1024 < 500:
        for i in range(0,len(created_jason_files_lake_riv)):
            injson2 = load(open(created_jason_files_lake_riv[i]))
            if 'finalcat_info_riv' in created_jason_files_lake_riv[i]:
                new_features = []
                for element in injson2["features"]:
                    if element["properties"]["Lake_Cat"] == 0:    
                        new_features.append(element) 
                injson2["features"] = new_features

            if i == 0:
                output_jason_lake_riv = injson2
            else:
                output_jason_lake_riv['features'] += injson2['features']
                
        with open(os.path.join(product_dir,'routing_product_lake_river.geojson'), 'w', encoding='utf-8') as f:
            json.dump(output_jason_lake_riv, f, ensure_ascii=False, indent=4)
    else:
        shutil.copy(created_jason_files_lake_riv[0], os.path.join(product_dir,'routing_product_lake_river.geojson'))             
    return 
    
##############


def qgis_raster_clip_raster_by_mask(processing, Input, MASK, TARGET_CRS, Output):
    """Functions clip raster by mask
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    out = processing.run(
        "gdal:cliprasterbymasklayer",
        {
            "INPUT": Input,
            "MASK": MASK,
            "SOURCE_CRS": None,
            "TARGET_CRS": QgsCoordinateReferenceSystem(TARGET_CRS),
            "NODATA": None,
            "ALPHA_BAND": False,
            "CROP_TO_CUTLINE": True,
            "KEEP_RESOLUTION": False,
            "SET_RESOLUTION": False,
            "X_RESOLUTION": None,
            "Y_RESOLUTION": None,
            "MULTITHREADING": False,
            "OPTIONS": "",
            "DATA_TYPE": 0,
            "EXTRA": "",
            "OUTPUT": Output,
        },
    )
    return out


def qgis_raster_clip_raster_by_extent(processing, Input, PROJWIN, Output):
    """Functions clip raster by extent
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """
    out = processing.run(
        "gdal:cliprasterbyextent",
        {
            "INPUT": Input,
            "PROJWIN": PROJWIN,
            "NODATA": None,
            "OPTIONS": "",
            "DATA_TYPE": 0,
            "EXTRA": "",
            "OUTPUT": Output,
        },
    )

    return out


def qgis_raster_saga_clip_raster_with_polygon(processing, context, Input, MASK, Output):
    """Functions clip raster by mask,
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    out = processing.run(
        "saga:cliprasterwithpolygon",
        {"INPUT": Input, "POLYGONS": MASK, "OUTPUT": Output},
        context=context,
    )

    return out


def qgis_raster_slope(processing, Input, Output):
    """Functions calculate slope from DEM
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    out = processing.run(
        "qgis:slope", {"INPUT": Input, "Z_FACTOR": 1, "OUTPUT": Output}
    )
    return out


def qgis_raster_aspect(processing, Input, Output):
    """Functions calculate aspect from DEM
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    out = processing.run(
        "qgis:aspect", {"INPUT": Input, "Z_FACTOR": 1, "OUTPUT": Output}
    )
    return out


def qgis_raster_zonal_statistics(processing, INPUT_RASTER, INPUT_VECTOR, COLUMN_PREFIX):
    """Functions calculate zonal statistics between polygon and raster
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    processing.run(
        "qgis:zonalstatistics",
        {
            "INPUT_RASTER": INPUT_RASTER,
            "RASTER_BAND": 1,
            "INPUT_VECTOR": INPUT_VECTOR,
            "COLUMN_PREFIX": COLUMN_PREFIX,
            "STATS": [2],
        },
    )
    return


def qgis_raster_read_raster(processing, INPUT):
    """qgis function read raster into memeory
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = QgsRasterLayer(INPUT, "")
    return out


def qgis_raster_gdal_translate(processing, INPUT, OUTPUT, format="GTiff"):
    """qgis function read raster into memeory
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    params = {"INPUT": INPUT, "format": format, "OUTPUT": OUTPUT}
    out = processing.run("gdal:translate", params)
    return out


def qgis_raster_return_raster_properties(processing, INPUT):
    """qgis function return raster projection and cellsize
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    cellSize = float(INPUT.rasterUnitsPerPixelX())  ### Get Raster cell size
    SpRef_in = INPUT.crs().authid()  ### get Raster spatialReference id
    return cellSize, SpRef_in


def qgis_raster_gdal_polygonize(processing, context, INPUT, OUTPUT):
    """qgis function polygonize rater to polygon
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run(
        "gdal:polygonize",
        {
            "INPUT": INPUT,
            "BAND": 1,
            "FIELD": "DN",
            "EIGHT_CONNECTEDNESS": False,
            "EXTRA": "",
            "OUTPUT": OUTPUT,
        },
        context=context,
    )

    return out


def qgis_raster_gdal_rasterize(
    processing, context, INPUT, Column_nm, cellsize, w, s, e, n, OUTPUT
):
    """qgis gdal function polygonize rater to polygon
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    cmd = (
        "gdal_rasterize -at -of GTiff -a_nodata -9999 -a "
        + Column_nm
        + " -tr  "
        + str(cellsize)
        + "  "
        + str(cellsize)
        + "  -te   "
        + str(w)
        + "   "
        + str(s)
        + "   "
        + str(e)
        + "   "
        + str(n)
        + "   "
        + '"'
        + INPUT
        + '"'
        + "    "
        + '"'
        + OUTPUT
        + '"'
    )
    os.system(cmd)
    return


# os.system('gdal_rasterize -at -of GTiff -a_nodata -9999 -a Hylak_id -tr  '+ str(self.cellSize) + "  " +str(self.cellSize)+'  -te   '+ str(grsregion['w'])+"   " +str(grsregion['s'])+"   " +str(grsregion['e'])+"   " +str(grsregion['n'])+"   " + "\"" +  self.Path_allLakeply +"\""+ "    "+ "\""+self.Path_allLakeRas+"\"")

def Copy_Pddataframe_to_shpfile_by_file(
    Path_shpfile,
    Pddataframe,
    link_col_nm_shp="SubId",
    link_col_nm_df="SubId",
    UpdateColNM=["#"],
    Input_Is_Feature_In_Mem=False,
):

    if Input_Is_Feature_In_Mem:
        layer_cat = Path_shpfile
    else:
        layer_cat = QgsVectorLayer(Path_shpfile, "")
    
    # table to csv 
    Attri_table.to_csv(os.path.join(tempfile.gettempdir(), "attribute.csv"),index = False)
    csv_uri = 'file:///%s?delimiter=,'%(os.path.join(tempfile.gettempdir(), "attribute.csv"))
    csv = QgsVectorLayer(csv_uri, 'attributes', 'delimitedtext')

    
    


def Copy_Pddataframe_to_shpfile(
    Path_shpfile,
    Pddataframe,
    link_col_nm_shp="SubId",
    link_col_nm_df="SubId",
    UpdateColNM=["#"],
    Input_Is_Feature_In_Mem=False,
):
    """Function modify attribute table of Path_shpfile using value from Pddataframe
    Parameters
    ----------
    Path_shpfile                        : shpfile
        Path to the shpfile
    Pddataframe                         : dataframe
        Dataframe constains data that will be used to update attribute table of
        Path_shpfile
    link_col_nm_shp                     : string
        The column name that link Pddataframe attribute table in Path_shpfile
    link_col_nm_df                      : string
        The column name that link Path_shpfile attribute table in Pddataframe
    UpdateColNM                         : list
        It is a list of column name, it is equal to '#', all column value will
        be updated, otherwise only column name in UpdateColNM will be updated.
    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """
    if Input_Is_Feature_In_Mem:
        layer_cat = Path_shpfile
    else:
        layer_cat = QgsVectorLayer(Path_shpfile, "")
        
    Attri_Name = layer_cat.fields().names()
    features = layer_cat.getFeatures()
    with edit(layer_cat):
        for sf in features:
            Atti_Valu = sf.attributes()
            sf_subid = sf[link_col_nm_shp]
            tarinfo = Pddataframe[Pddataframe[link_col_nm_df] == sf_subid]
            
            if UpdateColNM[0] == "#":
                for icolnm in range(0, len(Attri_Name)):  ### copy infomaiton
                    if (
                        Attri_Name[icolnm] == "Obs_NM"
                        or Attri_Name[icolnm] == "SRC_obs"
                        or Attri_Name[icolnm] == "LAND_USE_C"
                        or Attri_Name[icolnm] == "SOIL_PROF"
                        or Attri_Name[icolnm] == "VEG_C"
                    ):  
                        if str(tarinfo[Attri_Name[icolnm]].values[0]) == 'nan':
                            sf[Attri_Name[icolnm]] = qgis.core.NULL
                        else:
                            sf[Attri_Name[icolnm]] = str(
                                tarinfo[Attri_Name[icolnm]].values[0]
                            )
                    elif (
                        Attri_Name[icolnm] == "cat"
                        or Attri_Name[icolnm] == "layer"
                        or Attri_Name[icolnm] == "path"
                    ):
                        continue
                    else:
                        sf[Attri_Name[icolnm]] = float(
                            tarinfo[Attri_Name[icolnm]].values[0]
                        )
            else:
                for icolnm in range(0, len(UpdateColNM)):
                    if (
                        UpdateColNM[icolnm] == "Obs_NM"
                        or UpdateColNM[icolnm] == "SRC_obs"
                        or UpdateColNM[icolnm] == "LAND_USE_C"
                        or UpdateColNM[icolnm] == "SOIL_PROF"
                        or UpdateColNM[icolnm] == "VEG_C"
                    ):
                        sf[UpdateColNM[icolnm]] = str(
                            tarinfo[UpdateColNM[icolnm]].values[0]
                        )
                    elif (
                        UpdateColNM[icolnm] == "cat"
                        or UpdateColNM[icolnm] == "layer"
                        or UpdateColNM[icolnm] == "path"
                    ):
                        continue
                    else:
                        sf[UpdateColNM[icolnm]] = float(
                            tarinfo[UpdateColNM[icolnm]].values[0]
                        )

            layer_cat.updateFeature(sf)
    if Input_Is_Feature_In_Mem:
        return layer_cat
    else:
        del layer_cat
        return

def copy_data_and_dissolve(all_subids,tempfolder,processing,Path_Temp_final_rviply,Path_Temp_final_rvi,
    mapoldnew_info,COLUMN_NAMES_CONSTANT_CLEAN,OutputFolder,Path_Catchment_Polygon,context,Path_final_rviply_input = '#',Path_final_rvi_input = '#'):    
    
    if len(all_subids) > 5000:

        Path_final_rviply = os.path.join(
            tempfolder,
            "temp_finalriv_ply"  + ".shp",
        )
        Path_final_rvi = os.path.join(
            tempfolder,
            "temp_finalriv"  + ".shp",
        )
    
        Copy_Pddataframe_to_shpfile_main(
            Path_shpfile = Path_Temp_final_rviply,
            Pddataframe = mapoldnew_info,
            all_subids = all_subids,
            tempfolder = tempfolder,
            processing = processing,
            output = Path_final_rviply,
            link_col_nm_shp="SubId",
            link_col_nm_df="Old_SubId",
            UpdateColNM=["#"],
        )  
        
        Copy_Pddataframe_to_shpfile_main(
            Path_shpfile = Path_Temp_final_rvi,
            Pddataframe = mapoldnew_info,
            all_subids = all_subids,
            tempfolder = tempfolder,
            processing = processing,  
            output =   Path_final_rvi,    
            link_col_nm_shp="SubId",
            link_col_nm_df="Old_SubId",
            UpdateColNM=["#"],
        )    
                              
    else:
            
        Path_final_rviply = Path_Temp_final_rviply
        Path_final_rvi = Path_Temp_final_rvi
                    
        Copy_Pddataframe_to_shpfile(
            Path_Temp_final_rviply,
            mapoldnew_info,
            link_col_nm_shp="SubId",
            link_col_nm_df="Old_SubId",
            UpdateColNM=["#"],
        )
        Copy_Pddataframe_to_shpfile(
            Path_Temp_final_rvi,
            mapoldnew_info,
            link_col_nm_shp="SubId",
            link_col_nm_df="Old_SubId",
            UpdateColNM=["#"],
        )

    # dissolve shpfile based on new subid
    if len(os.path.basename(Path_Catchment_Polygon).split('_')) == 5:
        Path_final_rviply_out = os.path.join(OutputFolder, "finalcat_info_"+os.path.basename(Path_Catchment_Polygon).split('_')[4])
        Path_final_rvi_out = os.path.join(OutputFolder, "finalcat_info_riv_"+os.path.basename(Path_Catchment_Polygon).split('_')[4])
    else:
        Path_final_rviply_out = os.path.join(OutputFolder, "finalcat_info.shp")
        Path_final_rvi_out = os.path.join(OutputFolder, "finalcat_info_riv.shp")

    if Path_final_rviply_input != '#' and Path_final_rvi_input != '#':
        Path_final_rviply_out = os.path.join(OutputFolder, os.path.basename(Path_final_rviply_input))
        Path_final_rvi_out = os.path.join(OutputFolder, os.path.basename(Path_final_rvi_input))
        
    COLUMN_NAMES_CONSTANT_CLEAN_RIV = COLUMN_NAMES_CONSTANT_CLEAN.copy()
    COLUMN_NAMES_CONSTANT_CLEAN_RIV.remove('centroid_x') 
    COLUMN_NAMES_CONSTANT_CLEAN_RIV.remove('centroid_y')  
    Clean_Attribute_Name(Path_final_rvi, COLUMN_NAMES_CONSTANT_CLEAN_RIV)
    Clean_Attribute_Name(Path_final_rviply, COLUMN_NAMES_CONSTANT_CLEAN)
    
                    
    riv_dissolve = qgis_vector_dissolve(
        processing,
        context,
        INPUT=Path_final_rvi,
        FIELD=["SubId"],
        OUTPUT="memory:",# Path_final_rvi_out,
    )["OUTPUT"]

    processing.run("native:extractbyexpression", {
                    'INPUT':riv_dissolve,
                    'EXPRESSION':' \"Lake_Cat\" > 0  OR  \"RivLength\" > 0',
                    'OUTPUT':Path_final_rvi_out}
                    )

    ply_draft = qgis_vector_dissolve(
        processing,
        context,
        INPUT=Path_final_rviply,
        FIELD=["SubId"],
        OUTPUT="memory:",
    )["OUTPUT"]
    
    # clean attribute table of shpfile
    Clean_Attribute_Name(Path_final_rvi_out, COLUMN_NAMES_CONSTANT_CLEAN)
    ply_draft,tempnum = Clean_Attribute_Name(ply_draft, COLUMN_NAMES_CONSTANT_CLEAN,Input_Is_Feature_In_Mem = True)

    # add centroid to new drived polygons
    
    formular = "x(centroid(transform($geometry,'%s','%s')))" % (
        ply_draft.crs().authid(),
        "EPSG:4326",
    )

    ply_draft = qgis_vector_field_calculator(
        processing=processing,
        context=context,
        FORMULA=formular,
        FIELD_NAME="centroid_x",
        INPUT=ply_draft,
        OUTPUT="memory:",
    )["OUTPUT"]

    formular = "y(centroid(transform($geometry,'%s','%s')))" % (
        ply_draft.crs().authid(),
        "EPSG:4326",
    )

    qgis_vector_field_calculator(
        processing=processing,
        context=context,
        FORMULA=formular,
        FIELD_NAME="centroid_y",
        INPUT=ply_draft,
        OUTPUT=Path_final_rviply_out,
    )
    if 'finalcat_info' in Path_final_rviply_out:
        create_geo_jason_file(processing,Path_final_rviply_out)
    return 

def Copy_Pddataframe_to_shpfile_main(
    Path_shpfile,
    Pddataframe,
    all_subids,
    tempfolder,
    processing,
    output,
    link_col_nm_shp="SubId",
    link_col_nm_df="SubId",
    UpdateColNM=["#"],
    Input_Is_Feature_In_Mem=False,
):

    path_to_temp_files = []
    ncores = 3
    feature_per_file = int(len(all_subids)/ncores)
    
    for i in range(0,ncores):
        num_random = str(np.random.randint(1, 10000 + 1)) +'_'+str(i)
        path_of_i_temp_file = os.path.join(
            tempfolder,
            "temp_finalriv_ply" + num_random + ".shp",
        )
        if i != ncores - 1:
            subids_of_i_temp_file = all_subids[i*feature_per_file:(i+1)*feature_per_file]
        else:
            subids_of_i_temp_file = all_subids[i*feature_per_file:len(all_subids)]
                
        Selectfeatureattributes(
            processing,
            Input=Path_shpfile,
            Output=path_of_i_temp_file,
            Attri_NM="SubId",
            Values=subids_of_i_temp_file,
        )
        path_to_temp_files.append(path_of_i_temp_file)

    Parallel(n_jobs=ncores, verbose=1, backend="threading")(
        delayed(Copy_Pddataframe_to_shpfile)(temp_file,Pddataframe,
        link_col_nm_shp="SubId",link_col_nm_df="Old_SubId",UpdateColNM=["#"]
        ) for temp_file in path_to_temp_files)
    
    processing.run("native:mergevectorlayers", {'LAYERS':path_to_temp_files,'CRS':None,'OUTPUT':output})
    
    return 
    

def Remove_Unselected_Lake_Attribute_In_Finalcatinfo(Path_Finalcatinfo, Conn_Lake_Ids):
    """Functions will set lake id not in Conn_Lake_Ids to -1.2345 in attribute
        table of Path_Finalcatinfo
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    layer_cat = QgsVectorLayer(Path_Finalcatinfo, "")
    Attri_Name = layer_cat.fields().names()
    features = layer_cat.getFeatures()
    with edit(layer_cat):
        for sf in features:
            sf_subid = float(sf["HyLakeId"])

            if sf_subid in Conn_Lake_Ids or float(sf["Lake_Cat"]) == 2:
                continue
            sf["HyLakeId"] = float(0)
            sf["LakeVol"] = float(0)
            sf["LakeArea"] = float(0)
            sf["LakeDepth"] = float(0)
            sf["Laketype"] = float(0)
            sf["Lake_Cat"] = float(0)
            layer_cat.updateFeature(sf)
    del layer_cat
    return


#########
def Add_centroid_to_feature(
    Path_feagure, centroidx_nm="#", centroidy_nm="#", Input_Is_Feature_In_Mem=False
):
    """Functions will add centorid x y to Path_feagure
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """
    if Input_Is_Feature_In_Mem:
        layer_cat = Path_feagure
    else:
        layer_cat = QgsVectorLayer(Path_feagure, "")
    Attri_Name = layer_cat.fields().names()
    features = layer_cat.getFeatures()
    with edit(layer_cat):
        for sf in features:
            centroidxy = sf.geometry().centroid().asPoint()
            sf[centroidx_nm] = centroidxy[0]
            sf[centroidy_nm] = centroidxy[1]
            layer_cat.updateFeature(sf)
    if Input_Is_Feature_In_Mem:
        return layer_cat
    else:
        del layer_cat


def Selectfeatureattributes(processing, Input="#", Output="#", Attri_NM="#", Values=[],Is_str=False):
    """Functions extract features from Input, based on values in column Attri_NM
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """
    if Is_str:
        
        #'  \"STATION_NU\" IN (\'02KB001\',\'02KB003\')'
        
        exp = "\"%s\" IN (" %(Attri_NM)
        exp = exp + "\'%s\'" %(Values[0])
        for i in range(1, len(Values)):
            exp =  exp + " , \'%s\'" %(Values[i])
        exp = exp + ")"
    else:
        ncores = 3
        nvalues_per_group = 1000
        k = 0
        exp = Attri_NM + "  IN  (  " + str(int(Values[0]))
        for i in range(1, len(Values)):
            k = k + 1
            if k != 0:
                exp = exp + " , " + str(int(Values[i]))
            if k == nvalues_per_group and k < len(Values) - 3:
                k = 0
                exp = exp + ")"
                exp = exp + ' OR ' + Attri_NM + "  IN  (  " + str(int(Values[i]))
        
        exp = exp + ")"
    
    processing.run(
        "native:extractbyexpression",
        {"INPUT": Input, "EXPRESSION": exp, "OUTPUT": Output},
    )


def Copyfeature_to_another_shp_by_attribute(
    Source_shp, Target_shp, Col_NM="SubId", Values=[-1], Attributes=[-1]
):

    """Functions that will copy features in Source_shp to Target_shp
    based on attribute values in Values
    ----------

    Notes
    -------

    Returns:
    -------
        None, the attribute table of Path_shpfile will be updated
    """

    layer_src = QgsVectorLayer(Source_shp, "")
    layer_trg = QgsVectorLayer(Target_shp, "")

    src_features = layer_src.getFeatures()

    Selected_Features = []
    for sf in src_features:
        # centroidxy = sf.geometry().centroid().asPoint()
        Select_value = sf[Col_NM]
        if Select_value in Values:
            src_geometry = sf.geometry()
            attribute = Attributes.loc[Attributes[Col_NM] == Select_value].values
            temp_feature = QgsFeature()
            temp_feature.setGeometry(src_geometry)
            temp_feature.setAttributes(attribute.tolist()[0])
            Selected_Features.append(temp_feature)

    layer_trg.startEditing()
    layer_trg.addFeatures(Selected_Features)
    layer_trg.commitChanges()
    layer_trg.updateExtents()
    del layer_src
    del layer_trg


def Add_New_SubId_To_Subregion_shpfile(
    processing, context, Layer, SubID_info="#", OutputPath="#", Region_ID=1
):
    """Asign new subbasin Id to each subbasins in each subregion
    Parameters
    ----------
    processing                        : qgis object
    context                           : qgis object
    Layer                             : vector layer
        it is the subbasin polygon or polyline of watershed delineation
        result in each subregion
    SubID_info                        : dataframe
        it is a dataframe contains new subbasin id for each subbasin
        in Layer
    OutputPath                        : string
        Path to the output file
    Region_ID                         : integer
        it is the subregion id of layer

    Notes
    -------
        the output will be the same type of input Layer
        stored in OutputPath, a new SubId will be given
    Returns:
    -------
        None
    """

    layer2 = qgis_vector_field_calculator(
        processing=processing,
        context=context,
        FORMULA=str(int(Region_ID)),
        FIELD_NAME="Region_ID",
        INPUT=Layer,
        OUTPUT="memory:",
    )['OUTPUT']
    qgis_vector_field_calculator(
        processing=processing,
        context=context,
        FORMULA="0",
        FIELD_NAME="Use_region",
        INPUT=layer2,
        OUTPUT=OutputPath,
    )
    
    layer_new = QgsVectorLayer(OutputPath, "")

    features = layer_new.getFeatures()
    with edit(layer_new):
        for sf in features:
            cSubId = int(sf["SubId"])
            cDowSubId = int(sf["DowSubId"])
            nSubId = SubID_info.loc[SubID_info["SubId"] == cSubId]["nSubId"]
            if len(SubID_info.loc[SubID_info["SubId"] == cDowSubId]) == 0 or cSubId == cDowSubId:
                nDowSubId = -1
            else:
                nDowSubId = SubID_info.loc[SubID_info["SubId"] == cDowSubId]["nSubId"]
            nSeg_ID = SubID_info.loc[SubID_info["SubId"] == cSubId]["nSeg_ID"]
            sf["SubId"] = int(nSubId)
            sf["DowSubId"] = int(nDowSubId)
            sf["Seg_ID"] = int(nSeg_ID)
            layer_new.updateFeature(sf)
    del layer_new
    return


def qgis_vector_field_calculator(
    processing,
    context,
    FORMULA,
    INPUT,
    OUTPUT,
    FIELD_NAME,
    FIELD_PRECISION=0,
    FIELD_TYPE=0,
    NEW_FIELD=True,
    FIELD_LENGTH=10,
):
    """qgis filed calcuator
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    alg_params = {
        "FIELD_LENGTH": FIELD_LENGTH,
        "FIELD_NAME": FIELD_NAME,
        "FIELD_PRECISION": FIELD_PRECISION,
        "FIELD_TYPE": FIELD_TYPE,
        "FORMULA": FORMULA,
        "INPUT": INPUT,
        "NEW_FIELD": NEW_FIELD,
        "OUTPUT": OUTPUT,
    }

    out = processing.run("qgis:fieldcalculator", alg_params, context=context)
    return out


def qgis_vector_dissolve(
    processing, context, INPUT, FIELD, OUTPUT, USING_GDAL_FUNCTION=False
):
    """qgis dissolve input vector based on values in FIELD list
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    if USING_GDAL_FUNCTION:
        out = processing.run(
            "gdal:dissolve",
            {"INPUT": INPUT, "FIELD": FIELD, "OUTPUT": OUTPUT},
            context=context,
        )
    else:
        out = processing.run(
            "native:dissolve",
            {"INPUT": INPUT, "FIELD": FIELD, "OUTPUT": OUTPUT},
            context=context,
        )
    return out


def qgis_vector_fix_geometries(processing, context, INPUT, OUTPUT):
    """qgis fixgeometries
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run("native:fixgeometries", {"INPUT": INPUT, "OUTPUT": OUTPUT})
    return out


def qgis_vector_merge_vector_layers(processing, context, INPUT_Layer_List, OUTPUT):
    """qgis merge_vector_layers
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run(
        "native:mergevectorlayers", {"LAYERS": INPUT_Layer_List, "OUTPUT": OUTPUT}
    )

    return out


def qgis_vector_return_crs_id(
    processing, context, INPUT_Layer, Input_Is_Feature_In_Mem=True
):
    """qgis return vector layer projection crs id
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    if Input_Is_Feature_In_Mem:
        out = INPUT_Layer.crs().authid()
    else:
        layer = QgsVectorLayer(INPUT_Layer, "")
        out = layer.crs().authid()

    return out


def qgis_vector_add_polygon_attribute_to_points(
    processing, context, INPUT_Layer, POLYGONS, FIELDS, OUTPUT
):
    """qgis return vector layer projection crs id
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run(
        "saga:addpolygonattributestopoints",
        {
            "INPUT": INPUT_Layer,
            "POLYGONS": POLYGONS,
            "FIELDS": FIELDS,
            "OUTPUT": OUTPUT,
        },
    )

    return out


def qgis_vector_extract_by_attribute(
    processing, context, INPUT_Layer, FIELD, OPERATOR, VALUE, OUTPUT
):
    """qgis extract vector by attribute
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run(
        "native:extractbyattribute",
        {
            "INPUT": INPUT_Layer,
            "FIELD": FIELD,
            "OPERATOR": OPERATOR,
            "VALUE": VALUE,
            "OUTPUT": OUTPUT,
        },
    )

    return out


def qgis_vector_add_attributes(processing, context, INPUT_Layer, attribute_list):
    """qgis add attributes to vector
    ----------
    Notes
    -------

    Returns:
    -------
        None,
    """
    INPUT_Layer.dataProvider().addAttributes(attribute_list)
    INPUT_Layer.updateFields()
    INPUT_Layer.commitChanges()
    return INPUT_Layer


def qgis_vector_get_attributes(processing, context, INPUT_Layer, attribute_NM):
    """qgis retrun attribute value of vector
    ----------
    Notes
    -------

    Returns:
    -------
        None,
    """
    if attribute_NM == "count":
        out = INPUT_Layer.featureCount()
        return out
    elif attribute_NM == "field_name":
        out = INPUT_Layer.fields().names()
    elif attribute_NM == "features":
        out = INPUT_Layer.getFeatures()
        return out
    else:
        print("wrong attribute name")
        return -1


####


def qgis_vector_union_two_layers(
    processing, context, INPUT, OVERLAY, OUTPUT, OVERLAY_FIELDS_PREFIX=""
):
    """qgis union two layers
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """

    out = processing.run(
        "native:union",
        {
            "INPUT": INPUT,
            "OVERLAY": OVERLAY,
            "OVERLAY_FIELDS_PREFIX": OVERLAY_FIELDS_PREFIX,
            "OUTPUT": OUTPUT,
        },
        context=context,
    )

    return out


def qgis_vector_reproject_layers(processing, context, INPUT, TARGET_CRS, OUTPUT):
    """qgis function reproject vector layer
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run(
        "native:reprojectlayer",
        {
            "INPUT": INPUT,
            "TARGET_CRS": QgsCoordinateReferenceSystem(TARGET_CRS),
            "OUTPUT": OUTPUT,
        },
    )
    return out


def qgis_vector_buffer(processing, context, INPUT, Buffer_Distance, OUTPUT):
    """qgis function buffer vector layer with a buffer distance
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run(
        "native:buffer",
        {
            "INPUT": INPUT,
            "DISTANCE": Buffer_Distance,
            "SEGMENTS": 5,
            "END_CAP_STYLE": 0,
            "JOIN_STYLE": 0,
            "MITER_LIMIT": 2,
            "DISSOLVE": True,
            "OUTPUT": OUTPUT,
        },
        context=context,
    )
    return out


def qgis_vector_ectract_by_location(processing, context, INPUT, INTERSECT, OUTPUT):
    """qgis function extract vector by location
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run(
        "native:extractbylocation",
        {"INPUT": INPUT, "PREDICATE": [6], "INTERSECT": INTERSECT, "OUTPUT": OUTPUT},
        context=context,
    )
    return out


def qgis_vector_polygon_stro_lines(processing, context, INPUT, OUTPUT):
    """qgis function to obtain polygon boundary lines
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run(
        "native:polygonstolines", {"INPUT": INPUT, "OUTPUT": OUTPUT}, context=context
    )
    return out


def qgis_vector_clip(processing, context, INPUT, OVERLAY, OUTPUT):
    """qgis function reproject vector layer
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    layer_clip = processing.run(
        "native:clip", {"INPUT": INPUT, "OVERLAY": OVERLAY, "OUTPUT": OUTPUT}
    )

    return layer_clip


def qgis_vector_join_attribute_table(
    processing,
    context,
    INPUT1,
    FIELD1,
    INPUT2,
    FIELD2,
    OUTPUT,
    FIELDS_TO_COPY=[],
    METHOD=1,
    DISCARD_NONMATCHING=False,
    PREFIX="",
):
    """qgis function join attibute table
    ----------

    Notes
    -------

    Returns:
    -------
    """

    out = processing.run(
        "native:joinattributestable",
        {
            "INPUT": INPUT1,
            "FIELD": FIELD1,
            "INPUT_2": INPUT2,
            "FIELD_2": FIELD2,
            "FIELDS_TO_COPY": FIELDS_TO_COPY,
            "METHOD": METHOD,
            "DISCARD_NONMATCHING": DISCARD_NONMATCHING,
            "PREFIX": PREFIX,
            "OUTPUT": OUTPUT,
        },
        context=context,
    )
    #          processing.run("native:joinattributestable", {'INPUT':HRU_draft,'FIELD':Sub_ID,'INPUT_2':Path_Subbasin_Ply,'FIELD_2':Sub_ID,
    #                    'FIELDS_TO_COPY':[],'METHOD':1,'DISCARD_NONMATCHING':False,'PREFIX':'',
    #                    'OUTPUT':'memory:'},context = context)['OUTPUT']

    return out


def qgis_vector_create_spatial_index(processing, context, INPUT):
    """qgis function create spatial index
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = processing.run("qgis:createspatialindex", {"INPUT": INPUT})
    return out


def qgis_vector_read_vector(processing, context, INPUT):
    """qgis function read vector layer into memeory
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """
    out = QgsVectorLayer(INPUT, "")
    return out


def Obtain_Attribute_Table(
    processing, context, input_layer, Input_Is_Feature_In_Mem=True
):
    """Read QGIS vector layer attribute table

    Parameters
    ----------
    processing                        : qgis object
    context                           : qgis object
    vec_layer                         : qgis object
        a vector layer readed or generated by QGIS
    Returns:
    -------
    Attri_Tbl                  : Dataframe
       The attribute table in the input vector layer
    """
    
    if Input_Is_Feature_In_Mem:
        
        output_layer = os.path.join(tempfile.gettempdir(),'basinmaker_layer_'+str(np.random.randint(1, 10000 + 1))+".shp")
        
        _writer = QgsVectorFileWriter.writeAsVectorFormat(
        input_layer,
        output_layer,
        'utf-8',
        driverName='ESRI Shapefile'
        )
        vec_layer = output_layer

    else:
        vec_layer = input_layer
    
    Attri_Tbl = Dbf_To_Dataframe(vec_layer)

    # N_new_features = vec_layer.featureCount()
    # field_names = []
    # for field in vec_layer.fields():
    #     field_names.append(field.name())
    # Attri_Tbl = pd.DataFrame(
    #     data=np.full((N_new_features, len(field_names)), np.nan), columns=field_names
    # )
    # 
    # src_features = vec_layer.getFeatures()
    # i_sf = 0
    # for sf in src_features:
    #     for field in vec_layer.fields():
    #         filednm = field.name()
    #         Attri_Tbl.loc[Attri_Tbl.index[i_sf], filednm] = sf[filednm]
    #     i_sf = i_sf + 1
    return Attri_Tbl


def Clean_Attribute_Name(
    Input, FieldName_List, Input_Is_Feature_In_Mem=False, Col_NM_Max="SubId"
):
    """Function clean feature attribute table, all colnmun not in FieldName_List
        will be removed
    ----------

    Notes
    -------

    Returns:
    -------
        None,
    """

    fieldnames = set(FieldName_List)
    if Input_Is_Feature_In_Mem:
        layer_cat = Input
    else:
        layer_cat = QgsVectorLayer(Input, "")

    field_ids = []
    for field in layer_cat.fields():
        if field.name() not in fieldnames:
            field_ids.append(layer_cat.dataProvider().fieldNameIndex(field.name()))
        if field.name() == Col_NM_Max:
            max_subbasin_id = layer_cat.maximumValue(
                layer_cat.dataProvider().fieldNameIndex(field.name())
            )

    layer_cat.dataProvider().deleteAttributes(field_ids)
    layer_cat.updateFields()
    layer_cat.commitChanges()

    if Input_Is_Feature_In_Mem:
        return layer_cat, max_subbasin_id
    else:
        del layer_cat
        return max_subbasin_id

def change_neg_value_to_null_in_attribute_table(processing,Path_input_Layer,Path_output_layer):
        
    col_Name = 'Obs_NM'
    formular = 'CASE \r\n  WHEN \"%s\" =-9999 THEN NULL\r\n  ELSE \"%s\"\r\nEND' %(col_Name,col_Name)
    memout1 = processing.run("qgis:fieldcalculator", {
                   'INPUT':Path_input_Layer,
                   'FIELD_NAME':col_Name,
                   'FIELD_TYPE':2,
                   'FIELD_LENGTH':0,
                   'FIELD_PRECISION':0,
                   'FORMULA':formular,
                   'OUTPUT':"memory:",
                   })["OUTPUT"] 

    col_Name = 'SRC_obs'
    formular = 'CASE \r\n  WHEN \"%s\" =-9999 THEN NULL\r\n  ELSE \"%s\"\r\nEND' %(col_Name,col_Name)
    memout1 = processing.run("qgis:fieldcalculator", {
                   'INPUT':memout1,
                   'FIELD_NAME':col_Name,
                   'FIELD_TYPE':2,
                   'FIELD_LENGTH':0,
                   'FIELD_PRECISION':0,
                   'FORMULA':formular,
                   'OUTPUT':Path_output_layer,
                   })    
    return 
########



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import os


def rasterize_vectors_and_load_to_db(
    grassdb,
    grass_location,
    qgis_prefix_path,
    mask,
    vector_path,
    attribue_name,
    raster_name,
):

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    mask_layer = qgis_raster_read_raster(
        processing, os.path.join(grassdb, mask + ".tif")
    )  ### load DEM raster as a  QGIS raster object to obtain attribute
    cellSize, SpRef_in = qgis_raster_return_raster_properties(
        processing, mask_layer
    )  ### Get Raster cell size

    # load grass working location
    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    # get dem array and get nrows and ncols of the domain
    strtemp_array = Return_Raster_As_Array_With_garray(garray, mask)
    ncols = int(strtemp_array.shape[1])
    nrows = int(strtemp_array.shape[0])
    grsregion = gcore.region()

    qgis_raster_gdal_rasterize(
        processing,
        context,
        INPUT=vector_path,
        Column_nm=attribue_name,
        cellsize=cellSize,
        w=grsregion["w"],
        s=grsregion["s"],
        e=grsregion["e"],
        n=grsregion["n"],
        OUTPUT=os.path.join(grassdb, raster_name + ".tif"),
    )

    grass_raster_r_in_gdal(
        grass,
        raster_path=os.path.join(grassdb, raster_name + ".tif"),
        output_nm=raster_name,
    )

    grass_raster_setnull(
        grass, raster_nm=raster_name, null_values=[-9999], create_new_raster=False
    )

    grass_raster_v_import(grass, input_path=vector_path, output_vector_nm=raster_name)
    Qgs.exit()
    PERMANENT.close()


def obtain_lake_vectors_larger_than_threstholds(
    grassdb,
    qgis_prefix_path,
    all_lakes,
    lake_attributes,
    threshold_con_lake,
    threshold_non_con_lake,
    lakes_lg_cl_thres="lakes_lg_cl_thres",
    lakes_lg_ncl_thres="lakes_lg_ncl_thres",
):

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    print(threshold_con_lake, threshold_non_con_lake, lake_attributes[2])
    qgis_vector_extract_by_attribute(
        processing,
        context,
        INPUT_Layer=os.path.join(grassdb, all_lakes + ".shp"),
        FIELD=lake_attributes[2],
        OPERATOR=2,
        VALUE=threshold_con_lake,
        OUTPUT=os.path.join(grassdb, lakes_lg_cl_thres + ".shp"),
    )

    qgis_vector_extract_by_attribute(
        processing,
        context,
        INPUT_Layer=os.path.join(grassdb, all_lakes + ".shp"),
        FIELD=lake_attributes[2],
        OPERATOR=2,
        VALUE=threshold_non_con_lake,
        OUTPUT=os.path.join(grassdb, lakes_lg_ncl_thres + ".shp"),
    )
    Qgs.exit()



from basinmaker.func.arcgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def define_project_extent_using_dem(
    work_folder, 
    grass_location,
    qgis_prefix_path,
    path_dem_in, 
    mask="MASK", 
    dem="dem"
):

    """Define processing extent

    Function that used to define project processing spatial extent (PSE).
    The processing spatial extent is a region where Toolbox will work in. Toolbox
    will not process grids or features outside the processing spatial extent.
    Several options is available here. The PSE can be defined as the extent of
    input DEM.

    Parameters
    ----------
    grassdb                           : path (required)
        It is a path to project grass database folder
    grass_location                    : string (required)
        It is a string of grass location name
    qgis_prefix_path                  : string (required)
        It is a string of qgis prefix path
    path_dem_in                      : string (required)
        It is the path to input dem
    mask                             : string (optional)
        It is a output mask name, which will stored in grass_location in both
        vector and raster format
    dem                              : string (optional)
        It is a output dem raster name, which will be stored in grass_location

    Notes
    -------
    Outputs are following files

    MASK                   : raster
        it is a mask raster stored in grass database, which indicate
        the PSE. The grass database is located at
        os.path.join(grassdb, grass_location)
    dem                   : raster
        it is a dem raster stored in grass database, which is
        has the same extent with MASK. The grass database is located at
        os.path.join(grassdb, grass_location)

    Returns:
    -------
       None

    Examples
    -------
    """

    print("Mask Region:   Using provided DEM : ")
    
    # read and define arcgis work enviroments 
    arcpy.env.overwriteOutput = True
    arcpy.CheckOutExtension("Spatial")
    cellSize = float(arcpy.GetRasterProperties_management(path_dem_in, "CELLSIZEX").getOutput(0))
    SptailRef = arcpy.Describe(path_dem_in).spatialReference
    arcpy.env.XYTolerance = cellSize
    arcpy.arcpy.env.cellSize = cellSize    
    arcpy.env.outputCoordinateSystem = arcpy.SpatialReference(int(SptailRef.factoryCode)) ### WGS84
    if not os.path.exists(work_folder):
        os.makedirs(work_folder)
    arcpy.env.workspace = work_folder
    
    # copy dem to the work fokder 
    arcpy.CopyRaster_management(path_dem_in,dem + '.tif')
    arcpy.env.extent = arcpy.Describe(dem + '.tif').extent
    
    # create mask raster 
    mask_raster = Con(dem + '.tif', 1, 1, "VALUE >= 0")
    mask_raster.save(mask+'.tif')
    
    # create mask polygon 
    arcpy.RasterToPolygon_conversion(mask_raster, mask+'.shp', "NO_SIMPLIFY", "VALUE")

    return



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import os
from basinmaker.preprocessing.reprojectandclipvectorbyplyqgis import (
    reproject_clip_vectors_by_polygon,
)
from basinmaker.preprocessing.rasterizevectorsandloadtodbqgis import (
    rasterize_vectors_and_load_to_db,
)


def preprocessing_input_polygon(
    grassdb,
    grass_location,
    qgis_prefix_path,
    mask,
    path_polygon,
    ply_name,
    attribute_names=[],
    raster_names=[],
):

    reproject_clip_vectors_by_polygon(
        grassdb,
        grass_location,
        qgis_prefix_path,
        mask=os.path.join(grassdb, mask + ".shp"),
        path_polygon=path_polygon,
        ply_name=ply_name,
    )
    for i in range(0, len(attribute_names)):
        rasterize_vectors_and_load_to_db(
            grassdb,
            grass_location,
            qgis_prefix_path,
            mask,
            vector_path=os.path.join(grassdb, ply_name + ".shp"),
            attribue_name=attribute_names[i],
            raster_name=raster_names[i],
        )



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3
import pandas as pd
from basinmaker.preprocessing.preprocessrasterqgis import preprocess_raster


def calculate_flood_plain_manning_n(
    grassdb,
    grass_location,
    qgis_prefix_path,
    catinfo,
    input_geo_names,
    path_landuse="#",
    path_landuse_info="#",
):

    mask = input_geo_names["mask"]

    preprocess_raster(
        grassdb=grassdb,
        grass_location=grass_location,
        qgis_prefix_path=qgis_prefix_path,
        mask=mask,
        raster_path=path_landuse,
        raster_name="landuse",
    )

    cat_riv_info = input_geo_names["cat_riv_info"]
    outlet_pt_info = input_geo_names["outlet_pt_info"]

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    # read lanning's n and landuse type table
    landuse_and_n_table = pd.read_csv(path_landuse_info, sep=",")

    write_grass_reclass_rule_from_table(
        landuse_and_n_table.values, os.path.join(grassdb, "landuse_manning_rules.csv")
    )

    # viturally  landuse dataset

    grass_raster_r_external(
        grass, input=os.path.join(grassdb, "landuse_proj" + ".tif"), output="landuse_in"
    )
    # clip raster with mask in grass env
    grass_raster_r_clip(grass, input="landuse_in", output="landuse")
    # reclass landuse to manning's coefficient value *1000
    grass_raster_r_reclass(
        grass,
        input="landuse",
        output="landuse_Manning1",
        rules=os.path.join(grassdb, "landuse_manning_rules.csv"),
    )
    # calcuate real manning's coefficient for each landuse grid
    grass_raster_r_mapcalc(
        grass, expression="landuse_Manning = float(landuse_Manning1)/1000"
    )

    ### add averaged manning coefficent along the river network into river attribut table
    grass.run_command(
        "v.rast.stats",
        map=cat_riv_info,
        raster="landuse_Manning",
        column_prefix="mn",
        method=["average"],
    )

    grass.run_command(
        "v.what.rast", map=outlet_pt_info, raster= "landuse_Manning", column="mn_average"
    )
        
    # grass.run_command(
    #     "v.rast.stats",
    #     map=cat_ply_info,
    #     raster="landuse_Manning",
    #     column_prefix="mn",
    #     method=["average"],
    # )    

    ### read length and maximum and minimum dem along channel
    sqlstat = "SELECT Gridcode,mn_average FROM %s" % (cat_riv_info)
    rivleninfo = pd.read_sql_query(sqlstat, con)
    rivleninfo = rivleninfo.fillna(-9999)
    
    ### read length and maximum and minimum dem along channel
    sqlstat = "SELECT SubId,mn_average FROM %s" % (outlet_pt_info)
    outletpoint = pd.read_sql_query(sqlstat, con)
    outletpoint = outletpoint.fillna(-9999)

    for i in range(0, len(rivleninfo)):
        catid = rivleninfo["Gridcode"].values[i]
        catrow = catinfo["SubId"] == catid
        floodn = rivleninfo["mn_average"].values[i]
        
        if floodn < 0:
            floodn = outletpoint.loc[outletpoint['SubId'] ==catid, "mn_average"].values[0]
                        
        catinfo.loc[catrow, "FloodP_n"] = floodn

    PERMANENT.close()
    return catinfo



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def define_project_extent_using_input_polygon(
    grassdb,
    grass_location,
    qgis_prefix_path,
    path_dem_in,
    path_extent_ply,
    mask="MASK",
    dem="dem",
):

    """Define processing extent

    Function that used to define project processing spatial extent (PSE).
    The processing spatial extent is a region where Toolbox will work in. Toolbox
    will not process grids or features outside the processing spatial extent.
    Several options is available here. The PSE can be defined using input polygon

    Parameters
    ----------
    grassdb                           : path (required)
        It is a path to project grass database folder
    grass_location                    : string (required)
        It is a string of grass location name
    qgis_prefix_path                  : string (required)
        It is a string of qgis prefix path
    path_dem_in                      : string (required)
        It is the path to input dem
    path_extent_ply                  : string (optional)
        It is the path of a subregion polygon. It is only used when the Region
        of interest is very large and the resolution of the dem is very high.
        toolbox will first divide the whole region into several small subregions.
        And then using devided subregion polygon as PSE.
    mask                             : string (optional)
        It is a output mask name, which will stored in grass_location in both
        vector and raster format
    dem                              : string (optional)
        It is a output dem raster name, which will be stored in grass_location

    Notes
    -------
    Outputs are following files

    MASK                   : raster
        it is a mask raster stored in grass database, which indicate
        the PSE. The grass database is located at
        os.path.join(grassdb, grass_location)
    dem                   : raster
        it is a dem raster stored in grass database, which is
        has the same extent with MASK. The grass database is located at
        os.path.join(grassdb, grass_location)

    Returns:
    -------
       None

    Examples
    -------
    """

    print("mask region:   using input polygon ")

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )

    # load subregion mask polygon to target grass location
    PERMANENT = Session()
    PERMANENT.open(
        gisdb=grassdb,
        location=grass_location + "t1",
        create_opts="EPSG:4326",
    )
    grass_raster_v_in_org(
        grass,
        input_path=path_extent_ply,
        output_vector_nm=mask,
        location=grass_location,
    )
    PERMANENT.close()

    # unpack dem and generate mask with loaded vector

    PERMANENT = Session()
    # grass unpack function report an error if gisdb is provided using relative
    # path, so absolute path is provided here
    PERMANENT.open(
        gisdb=os.path.abspath(grassdb), location=grass_location, create_opts=""
    )

    # unpack dem dataset
    root, ext = os.path.splitext(path_dem_in)
    if ext == ".pack":
        grass_raster_r_unpack(
            grass, input=os.path.abspath(path_dem_in), output=dem + "_big"
        )
    else:
        grass_raster_r_in_gdal(
            grass=grass, raster_path=path_dem_in, output_nm=dem + "_big"
        )
    ###Use polygon to define a smaller region
    # use big dem define region first and then change it by vector mask
    grass_raster_g_region(grass, dem + "_big")
    # define mask of current working enviroments
    grass_raster_r_mask(grass, raster_nm="#", vector_nm=mask)
    # define new region with this mask
    grass_raster_g_region(grass, raster_nm="#", zoom="MASK")

    # using vector define mask and clip big raster dem
    grass_raster_r_mask(grass, raster_nm="#", vector_nm=mask)
    exp = "%s = %s" % (dem, dem + "_big")
    grass_raster_r_mapcalc(grass, expression=exp)

    # using clipped dem to defing mask
    grass_raster_r_mask(grass, raster_nm=dem)

    # export dem and mask and generate mask polygon
    grass_raster_r_out_gdal(
        grass,
        input_nm=mask,
        output=os.path.join(grassdb, mask + ".tif"),
        format="GTiff",
    )
    grass_raster_r_out_gdal(
        grass,
        input_nm=dem,
        output=os.path.join(grassdb, dem + ".tif"),
        format="GTiff",
    )
    # polygonize exported mask raster, and polygonize and dissove it.
    qgis_raster_gdal_polygonize(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + ".tif"),
        OUTPUT=os.path.join(grassdb, mask + "1.shp"),
    )
    # qgis dissolve function for some reason not working here....
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=os.path.join(grassdb, mask + "1.shp"),
        FIELD="DN",
        OUTPUT=os.path.join(grassdb, mask + ".shp"),
        USING_GDAL_FUNCTION=True,
    )

    PERMANENT.close()

    Qgs.exit()



import os
import geopandas
import matplotlib.pyplot as plt ## only needed to plot figures
from ipywidgets import HTML,Layout,IntSlider, ColorPicker, jslink ## only needed to plot figures
from ipyleaflet import Map, GeoData, basemaps, LayersControl,Popup,Marker,Polygon,Choropleth,WidgetControl## only needed to plot figures
#import leafmap.foliumap  as leafmap

def plot_routing_product_with_ipyleaflet(path_to_product_folder,version_number = ''):
    product_folder = path_to_product_folder
    if version_number != '':
        version_number = '_'+version_number
    path_subbasin = os.path.join(product_folder,'finalcat_info'+version_number+'.shp')
    path_river = os.path.join(product_folder,'finalcat_info_riv'+version_number+'.shp')
    path_cllake = os.path.join(product_folder,'sl_connected_lake'+version_number+'.shp')
    path_ncllake = os.path.join(product_folder,'sl_non_connected_lake'+version_number+'.shp')

    subbasin = geopandas.read_file(path_subbasin)
    subbasin = subbasin.to_crs("EPSG:4326")

    if os.path.exists(path_river):
        river = geopandas.read_file(path_river)
        river = river.to_crs("EPSG:4326")


    if os.path.exists(path_cllake):
        cllake = geopandas.read_file(path_cllake)
        cllake = cllake.to_crs("EPSG:4326")
    if os.path.exists(path_ncllake):
        ncllake = geopandas.read_file(path_ncllake)
        ncllake = ncllake.to_crs("EPSG:4326")

    cx = subbasin['centroid_x'].mean()
    cy = subbasin['centroid_y'].mean()
    m = Map(center=(cy,cx), zoom = 9, basemap= basemaps.Esri.WorldTopoMap)

    subnolake = subbasin[subbasin['Lake_Cat'] == 0].copy(deep=True)
    sub_nolake_map = GeoData(geo_dataframe = subnolake,
                      style={'color': '#6E6E6E', 'fillColor': '#89CD66', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1},
                      name = 'Subbasin without lakes')
    subcllake = subbasin[subbasin['Lake_Cat'] == 1].copy(deep=True)
    if len(subcllake) > 0:
        sub_cllake_map = GeoData(geo_dataframe = subcllake,
                          style={'color': '#6E6E6E', 'fillColor': '#CDE389', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1},
                          name = 'Subbasin with connected lakes')
    subncllake = subbasin[subbasin['Lake_Cat'] == 2].copy(deep=True)
    if len(subncllake) > 0:
        sub_ncllake_map = GeoData(geo_dataframe = subncllake,
                          style={'color': '#6E6E6E', 'fillColor': '#F5F57A', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1},
                          name = 'Subbasin with connected lakes')
    if os.path.exists(path_river):
        river_map = GeoData(geo_dataframe = river,
                          style={'color': '#0070FF', 'fillColor': '#0070FF', 'opacity':1, 'weight':2, 'dashArray':'2', 'fillOpacity':1},
                          name = 'River')
    if os.path.exists(path_cllake):
        cllake_map = GeoData(geo_dataframe = cllake,
                            style={'color': '#6E6E6E', 'fillColor': '#0070FF', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1},
                            name = 'Non connected lakes')
    if os.path.exists(path_ncllake):
        ncllake_map = GeoData(geo_dataframe = ncllake,
                             style={'color': '#6E6E6E', 'fillColor': '#0070FF', 'opacity':0, 'weight':1, 'dashArray':'2', 'fillOpacity':1},
                             name = 'Connected lakes')

    m.add_layer(sub_nolake_map)
    if len(subcllake) > 0:
        m.add_layer(sub_cllake_map)
    if len(subncllake) > 0:
        m.add_layer(sub_ncllake_map)
    if os.path.exists(path_river):
        m.add_layer(river_map)
    if os.path.exists(path_cllake):
        m.add_layer(cllake_map)
    if os.path.exists(path_ncllake):
        m.add_layer(ncllake_map)

    m.add_control(LayersControl())
    m.layout.height="700px"

    return m

# def plot_routing_product_with_leafmap(path_to_product_folder,version_number = ''):
#     product_folder = path_to_product_folder
#     if version_number != '':
#         version_number = '_'+version_number
#     path_subbasin = os.path.join(product_folder,'finalcat_info'+version_number+'.shp')
#     path_river = os.path.join(product_folder,'finalcat_info_riv'+version_number+'.shp')
#     path_cllake = os.path.join(product_folder,'sl_connected_lake'+version_number+'.shp')
#     path_ncllake = os.path.join(product_folder,'sl_non_connected_lake'+version_number+'.shp')
#
#     subbasin = geopandas.read_file(path_subbasin)
#     subbasin = subbasin.to_crs("EPSG:4326")
#
#     if os.path.exists(path_river):
#         river = geopandas.read_file(path_river)
#         river = river.to_crs("EPSG:4326")
#
#
#     if os.path.exists(path_cllake):
#         cllake = geopandas.read_file(path_cllake)
#         cllake = cllake.to_crs("EPSG:4326")
#     if os.path.exists(path_ncllake):
#         ncllake = geopandas.read_file(path_ncllake)
#         ncllake = ncllake.to_crs("EPSG:4326")
#
#     subcol_olrrp = ['SubId','geometry','DowSubId','BasArea','Lake_Cat','DrainArea','DA_Chn_L','DA_Slope','DA_Chn_Slp','BasSlope']
#     subcol_na = ['SubId','geometry','DowSubId','BasArea','Lake_Cat','DrainArea','BasSlope']
#
#     rivcol = ['geometry','RivSlope','RivLength','BkfWidth','BkfDepth','Strahler','Q_Mean','FloodP_n','Ch_n']
#     lake_col =  ['geometry','Hylak_id','Lake_type','Lake_area','Vol_total','Depth_avg']
#
#     if version_number == 'v1-0':
#         subcol = subcol_olrrp
#     else:
#         subcol = subcol_na
#
#     subnolake = subbasin[subbasin['Lake_Cat'] == 0].copy(deep=True)
#
#     m = leafmap.Map(
#         draw_control=False,
#         measure_control=False,
#         fullscreen_control=False,
#         attribution_control=True,
#     )
#
#
#     m.add_gdf(subnolake[subcol], layer_name="Subbasin without lakes",style = {'color': '#6E6E6E', 'fillColor': '#89CD66', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1})
#     labels = ["Subbasin without lakes"]
#     colors = ["#89CD66"]
#
#     subcllake = subbasin[subbasin['Lake_Cat'] == 1].copy(deep=True)
#     if len(subcllake) > 0:
#         style = {'color': '#6E6E6E', 'fillColor': '#CDE389', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1}
#         name = 'Subbasin with connected lakes'
#         m.add_gdf(subcllake[subcol], layer_name=name,style = style)
#         labels = labels + ['Subbasin with connected lakes']
#         colors = colors + ['#CDE389']
#
#
#     subncllake = subbasin[subbasin['Lake_Cat'] == 2].copy(deep=True)
#     if len(subncllake) > 0:
#         style = {'color': '#6E6E6E', 'fillColor': '#F5F57A', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1}
#         name = 'Subbasin with non connected lakes'
#         m.add_gdf(subncllake[subcol], layer_name=name,style = style)
#         labels = labels + ['Subbasin with non connected lakes']
#         colors = colors + ['#F5F57A']
#
#     if os.path.exists(path_river):
#         style = {'color': '#0070FF', 'fillColor': '#0070FF', 'opacity':1, 'weight':2, 'dashArray':'2', 'fillOpacity':1}
#         name = 'River'
#         m.add_gdf(river[rivcol], layer_name=name,style = style)
#         labels = labels + ['River/Lakes']
#         colors = colors + ['#0070FF']
#
#     if os.path.exists(path_cllake):
#         style = {'color': '#6E6E6E', 'fillColor': '#0070FF', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1}
#         name = 'Non connected lakes'
#         m.add_gdf(cllake[lake_col], layer_name=name,style = style)
#
#
#     if os.path.exists(path_ncllake):
#         style = {'color': '#6E6E6E', 'fillColor': '#0070FF', 'opacity':1, 'weight':1, 'dashArray':'2', 'fillOpacity':1}
#         name = 'Cconnected lakes'
#         m.add_gdf(ncllake[lake_col], layer_name=name,style = style)
#
#     m.add_legend(title='Legend', labels=labels, colors=colors)
#     return m



import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
from postprocessing.increasedaarcgis import simplify_routing_structure_by_drainage_area_arcgis
from postprocessing.combinearcgis import combine_catchments_covered_by_the_same_lake_arcgis

Routing_Product_Folder = sys.argv[1]
Area_Min=float(sys.argv[2])
OutputFolder =sys.argv[3]


simplify_routing_structure_by_drainage_area_arcgis(
    Routing_Product_Folder = Routing_Product_Folder,
    Area_Min=Area_Min,
    OutputFolder=OutputFolder,
)
        

combine_catchments_covered_by_the_same_lake_arcgis(
    Routing_Product_Folder=OutputFolder,
)  


import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
from postprocessing.selectprodarcgis import Select_Routing_product_based_SubId_arcgis
from postprocessing.combinearcgis import combine_catchments_covered_by_the_same_lake_arcgis

OutputFolder =sys.argv[4]
mostdownid_in=sys.argv[2]
mostupstreamid_in=sys.argv[3]
Routing_Product_Folder = sys.argv[1]

arcpy.AddMessage(Routing_Product_Folder)

mostdownid_in = mostdownid_in.split(";")
mostdownid = []
for i in range(0,len(mostdownid_in)):
    try:
        mostdownid.append(int(mostdownid_in[i]))
    except:
        mostdownid = []
        arcpy.AddMessage("Invalid input most down stream subbasin IDs  ", mostdownid_in)
        exist()

mostupstreamid_in = mostupstreamid_in.split(";")
        
mostupid = []
for i in range(0,len(mostupstreamid_in)):
    try:
        mostupid.append(int(mostupstreamid_in[i]))
    except:
        mostupid = []
        arcpy.AddMessage("Invalid input most up stream subbasin IDs  ", mostupid)
        exist()

arcpy.AddMessage(mostdownid)
arcpy.AddMessage(mostupid)
        
Select_Routing_product_based_SubId_arcgis(
    OutputFolder=OutputFolder,
    mostdownid=mostdownid,
    mostupstreamid=mostupid,
    Routing_Product_Folder = Routing_Product_Folder,
)
        

combine_catchments_covered_by_the_same_lake_arcgis(
    Routing_Product_Folder=OutputFolder,
)  



import os
import wget
import pandas as pd
import numpy as np
import gdown
# define function two download routing product for a given gauge
def Download_Routing_Product_For_One_Gauge(gauge_name,product_name,region='#',subreg = '#'):
    if product_name == 'NALRP':
        version = 'v2-1'
        if gauge_name != '#':
            gauge_info = pd.read_csv("https://github.com/dustming/RoutingTool/wiki/Files/obs_gauges_NA_v2-1.csv")
            gauge_info_sl = gauge_info[gauge_info['Obs_NM'] == gauge_name]
            if len(gauge_info_sl) < 1:
                print("The gauge ",gauge_name,"is not included in the routing product ")
                SubId = -1
                product_path = '#'
            else:
                if gauge_info_sl['Use_region'].values[0] < 1:
#                    print(gauge_info_sl)
#                    print(gauge_info_sl['Region'].values[0],gauge_info_sl['Sub_Reg'].values[0])
                    region_id = int(gauge_info_sl['Region'].values[0])
                    subreg_id = int(gauge_info_sl['Sub_Reg'].values[0])
                    url = "http://hydrology.uwaterloo.ca/basinmaker/data/original/drainage_region_%s_%s_v2-1.zip" % (str(region_id).zfill(4),str(subreg_id).zfill(5))
                    wget.download(url)
                    os.system('unzip drainage_region_%s_%s_%s.zip' % (str(region_id).zfill(4),str(subreg_id).zfill(5),version))
                    SubId =  gauge_info_sl['SubId'].values[0]
                    product_name = "drainage_region_%s_%s_%s" % (str(region_id).zfill(4),str(subreg_id).zfill(5),version)
                    product_path = os.path.join(os.getcwd(),product_name)
                    print("The needed product locates at:",product_path)
                    print("The Subbasin Id of the interested gauge is:",SubId)
                else:
#                    print(gauge_info_sl)
#                    print(gauge_info_sl['Region'].values[0],gauge_info_sl['Sub_Reg'].values[0])
                    region_id = int(gauge_info_sl['Region'].values[0])
#                    subreg_id = int(gauge_info_sl['Sub_Reg'].values[0])
                    url = "http://hydrology.uwaterloo.ca/basinmaker/data/original/drainage_region_%s_v2-1.zip" % (str(region_id).zfill(4))
                    wget.download(url)
                    os.system('unzip drainage_region_%s_v2-1.zip' % (str(region_id).zfill(4)))
                    SubId =  gauge_info_sl['SubId'].values[0]
                    product_name = "drainage_region_%s_v2-1" % (str(region_id).zfill(4))
                    product_path = os.path.join(os.getcwd(),product_name)
                    print("The needed product locates at:",product_path)
                    print("The Subbasin Id of the interested gauge is:",SubId)

        elif region != '#' and subreg =='#':
            region_id = region
            subreg_id = '-'
            url = "http://hydrology.uwaterloo.ca/basinmaker/data/original/drainage_region_%s_v2-1.zip" % (str(region_id).zfill(4))
            wget.download(url)
            os.system('unzip drainage_region_%s_v2-1.zip' % (str(region_id).zfill(4)))
            product_name =  "drainage_region_%s_v2-1" % (str(region_id).zfill(4))
            product_path = os.path.join(os.getcwd(),product_name)
            SubId =  -1

        elif region != '#' and subreg !='#':
            print("todo")
        else:
            print("Wrong option provided")


    if product_name == 'OLRP':
        version = 'v1-0'
        gauge_info = pd.read_csv("https://github.com/dustming/RoutingTool/wiki/Files/OIH_gauge_info.csv")
        productinfo = pd.read_csv("https://github.com/dustming/RoutingTool/wiki/Files/OLRRP.csv")
        if gauge_name != '#':
            gauge_info_sl = gauge_info[gauge_info['Obs_NM'] == gauge_name]
            if len(gauge_info_sl) < 1 or gauge_info_sl['SubId'].values[0] < 0:
                print("The gauge ",gauge_name,"is not included in the routing product ")
                SubId = -1
                product_path = '#'
            else:
                if gauge_info_sl['Use_region'].values[0] < 1:
                    region_id = gauge_info_sl['Region'].values[0]
                    subreg_id = gauge_info_sl['Sub_Reg'].values[0]
                    mask1 = productinfo['Region'] == region_id
                    mask2 = productinfo['Sub_Reg'] == subreg_id
                    mask = np.logical_and(mask1,mask2)
                    url_veiw = productinfo.loc[mask,'Download'].values[0]

                    url_veiw = url_veiw.split("/")
                    url = "https://drive.google.com/u/0/uc?id=%s&export=download"%(url_veiw[5])
                    output = 'drainage_region_%s_%s_%s.zip' % (region_id,subreg_id,version)

                    os.system('gdown %s -O %s' % (url_veiw[5],output))

                    os.system('unzip drainage_region_%s_%s_%s.zip' % (region_id,subreg_id,version))
                    SubId =  gauge_info_sl['SubId'].values[0]
                    product_name = 'drainage_region_%s_%s_%s' % (region_id,subreg_id,version)
                    product_path = os.path.join(os.getcwd(),product_name)
                    print("The needed product locates at:",product_path)
                    print("The Subbasin Id of the interested gauge is:",SubId)
                else:
                    region_id = gauge_info_sl['Region'].values[0]
                    subreg_id = '-'
                    mask1 = productinfo['Region'] == region_id
                    mask2 = productinfo['Sub_Reg'] == subreg_id
                    mask = np.logical_and(mask1,mask2)
                    url_veiw = productinfo.loc[mask,'Download'].values[0]
                    url_veiw = url_veiw.split("/")
                    url = "https://drive.google.com/u/0/uc?id=%s&export=download"%(url_veiw[5])
                    output = 'drainage_region_%s_%s.zip' % (region_id,version)

                    os.system('gdown %s -O %s' % (url_veiw[5],output))

                    os.system('unzip drainage_region_%s_%s.zip' % (region_id,version))
                    SubId =  gauge_info_sl['SubId'].values[0]
                    product_name ='drainage_region_%s_%s' % (region_id,version)
                    product_path = os.path.join(os.getcwd(),product_name)
                    print("The needed product locates at:",product_path)
                    print("The Subbasin Id of the interested gauge is:",SubId)

        elif region != '#' and subreg =='#':
            region_id = region
            subreg_id = '-'
            mask1 = productinfo['Region'] == region_id
            mask2 = productinfo['Sub_Reg'] == subreg_id
            mask = np.logical_and(mask1,mask2)
            url_veiw = productinfo.loc[mask,'Download'].values[0]
            url_veiw = url_veiw.split("/")
            url = "https://drive.google.com/u/0/uc?id=%s&export=download"%(url_veiw[5])
            output = 'drainage_region_%s_%s.zip' % (region_id,version)

            os.system('gdown %s -O %s' % (url_veiw[5],output))

            os.system('unzip drainage_region_%s_%s.zip' % (region_id,version))
            SubId =  -1
            product_name ='drainage_region_%s_%s' % (region_id,version)
            product_path = os.path.join(os.getcwd(),product_name)

        elif region != '#' and subreg !='#':
            print("todo")
        else:
            print("Wrong option provided")

    return SubId,product_path



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
from basinmaker.preprocessing.preprocessinglakeply import preprocessing_lake_polygon
import sqlite3
from basinmaker.addlakeandobs.definelaketypeqgis import (
    define_connected_and_non_connected_lake_type,
)
from basinmaker.addlakeandobs.filterlakesqgis import select_lakes_by_area_r
from basinmaker.addlakeandobs.pourpointsqgis import define_pour_points_with_lakes
from basinmaker.addlakeandobs.modifyfdr import modify_lakes_flow_direction
from basinmaker.addlakeandobs.modifystr import modify_str_r_to_add_sub_in_head_str


def add_lakes_into_existing_watershed_delineation(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    path_lakefile_in,
    lake_attributes,
    threshold_con_lake,
    threshold_non_con_lake,
    only_included_lake_at_river_interction = False,
    remove_lake_inlets = False,
    path_sub_reg_lake_r="#",
    path_sub_reg_lake_bd_r="#",
    sl_connected_lake="sl_connected_lake",
    sl_non_connected_lake="sl_nonconnect_lake",
    sl_lakes="selected_lakes",
    sl_str_connected_lake="str_sl_connected_lake",
    nfdr_arcgis="narcgis_fdr",
    nfdr_grass="ngrass_fdr",
    cat_add_lake="cat_add_lake",
    pourpoints_with_lakes="pourpoints_with_lakes",
    cat_use_default_acc="cat_use_default_acc",
    lake_outflow_pourpoints="lake_outflow_pourpoints",
    problem_seg="problem_seg",
    max_memroy=1024 * 4,
):
    # define required input files names
    fdr_arcgis = input_geo_names["fdr_arcgis"]
    fdr_grass = input_geo_names["fdr_grass"]
    str_r = input_geo_names["str_r"]
    str_v = input_geo_names["str_v"]
    acc = input_geo_names["acc"]
    cat_no_lake = input_geo_names["cat_no_lake"]
    mask = input_geo_names["mask"]
    dem = input_geo_names["dem"]

    # define internal file names
    lake_inflow_pourpoints = Internal_Constant_Names["lake_inflow_pourpoints"]
    catchment_pourpoints_outside_lake = Internal_Constant_Names[
        "catchment_pourpoints_outside_lake"
    ]
    cat_add_lake_old_fdr = Internal_Constant_Names["cat_add_lake_old_fdr"]
    str_connected_lake = Internal_Constant_Names["str_connected_lake"]
    alllake = Internal_Constant_Names["all_lakes"]
    lake_boundary = Internal_Constant_Names["lake_boundary"]
    connected_lake = Internal_Constant_Names["connect_lake"]
    non_connected_lake = Internal_Constant_Names["nonconnect_lake"]
    lakes_lg_cl_thres = "lakes_lg_cl_thres"
    lakes_lg_ncl_thres = "lakes_lg_ncl_thres"
    # prepropessing lakes inputs
    if path_lakefile_in == "#":
        import grass.script as grass
        import grass.script.setup as gsetup
        from grass.pygrass.modules import Module
        from grass.pygrass.modules.shortcuts import general as g
        from grass.pygrass.modules.shortcuts import raster as r
        from grass.script import array as garray
        from grass.script import core as gcore
        from grass_session import Session

        os.environ.update(
            dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
        )
        PERMANENT = Session()
        PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

        con = sqlite3.connect(
            os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
        )

        routing_info = generate_routing_info_of_catchments(
            grass, con, cat=cat_no_lake, acc=acc, str=str_r, Name="cat1",garray=garray
        )
        grass.run_command(
            "g.copy", rast=("cat1_OL", pourpoints_with_lakes), overwrite=True
        )

        lake_outflow_pourpoints = '#'
        grass.run_command(
            "g.copy", rast=(cat_no_lake, cat_use_default_acc), overwrite=True
        )
        grass.run_command(
            "g.copy", rast=(cat_no_lake, cat_add_lake), overwrite=True
        )
        grass.run_command("g.copy", rast=(fdr_grass, nfdr_grass), overwrite=True)
        grass.run_command("g.copy", rast=(fdr_arcgis, nfdr_arcgis), overwrite=True)
        return lake_outflow_pourpoints

    preprocessing_lake_polygon(
        path_lakefile_in=path_lakefile_in,
        lake_attributes=lake_attributes,
        mask=mask,
        grassdb=grassdb,
        grass_location=grass_location,
        qgis_prefix_path=qgis_prefix_path,
        threshold_con_lake=threshold_con_lake,
        threshold_non_con_lake=threshold_non_con_lake,
        gis_platform="qgis",
        lake_name=alllake,
        lake_boundary_name=lake_boundary,
        lakes_lg_cl_thres=lakes_lg_cl_thres,
        lakes_lg_ncl_thres=lakes_lg_ncl_thres,
    )

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    if path_sub_reg_lake_r != "#":
        grass.run_command(
            "r.unpack", input=path_sub_reg_lake_r, output="rg_lake", overwrite=True
        )
        exp = "%s = int(%s)" % ("rg_lake", "rg_lake")
        grass_raster_r_mapcalc(grass, exp)
    
        grass.run_command(
            "r.unpack",
            input=path_sub_reg_lake_bd_r,
            output="rg_lake_bd",
            overwrite=True,
        )
        exp = "%s = int(%s)" % ("rg_lake_bd", "rg_lake_bd")
        grass_raster_r_mapcalc(grass, exp)
    
        exp = "%s = if(isnull(%s),%s,%s)" % (alllake,"rg_lake",alllake,"rg_lake")
        grass_raster_r_mapcalc(grass, exp)
        exp = "%s = if(isnull(%s),%s,%s)" %  (lake_boundary,"rg_lake_bd",lake_boundary,"rg_lake")
        exp = "%s = int(%s)" % (alllake, alllake)
        grass_raster_r_mapcalc(grass, exp)
        exp = "%s = int(%s)" % (lake_boundary,lake_boundary)
        grass_raster_r_mapcalc(grass, exp)
    
    else:
        write_grass_and_arcgis_fdr_rules(grassdb)
        exp = "%s = int(%s)" % (alllake, alllake)
        grass_raster_r_mapcalc(grass, exp)
        exp = "%s = int(%s)" % (lake_boundary, lake_boundary)
        grass_raster_r_mapcalc(grass, exp)
    
    # Define connected and non connected lakes and
    # identify which str make certain lake have two outlet
    define_connected_and_non_connected_lake_type(
        grass,
        con,
        garray,
        str_r=str_r,
        lake=alllake,
        connected_lake=connected_lake,
        non_connected_lake=non_connected_lake,
        str_connected_lake=str_connected_lake,
    )
    
    select_lakes_by_area_r(
        grass=grass,
        con=con,
        garray=garray,
        str_r=str_r,
        threshold_con_lake=threshold_con_lake,
        threshold_non_con_lake=threshold_non_con_lake,
        only_included_lake_at_river_interction = only_included_lake_at_river_interction,
        lake_attributes=lake_attributes,
        lakes=alllake,
        connected_lake=connected_lake,
        non_connected_lake=non_connected_lake,
        str_connected_lake=str_connected_lake,
        sl_connected_lake=sl_connected_lake,
        sl_non_connected_lake=sl_non_connected_lake,
        sl_lakes=sl_lakes,
        sl_str_connected_lake=sl_str_connected_lake,
    )
    
    modify_str_r_to_add_sub_in_head_str(
        grass = grass,
        garray = garray,
        con = con,
        str_r = str_r,
        str_v = str_v,
        connected_lake = connected_lake,
        cat_no_lake = cat_no_lake,
        fdr_grass = fdr_grass,
        max_memroy = max_memroy,      
    )

    Lakes_WIth_Multi_Outlet, Remove_Str = define_pour_points_with_lakes(
        grass=grass,
        con=con,
        garray=garray,
        str_r=str_r,
        remove_lake_inlets = remove_lake_inlets,
        cat_no_lake=cat_no_lake,
        sl_lakes=sl_lakes,
        sl_connected_lake=sl_connected_lake,
        sl_str_connected_lake=sl_str_connected_lake,
        acc=acc,
        pourpoints_with_lakes=pourpoints_with_lakes,
        lake_inflow_pourpoints=lake_inflow_pourpoints,
        lake_outflow_pourpoints=lake_outflow_pourpoints,
        catchment_pourpoints_outside_lake=catchment_pourpoints_outside_lake,
    )

    grass.run_command(
        "r.stream.basins",
        direction=fdr_grass,
        points=pourpoints_with_lakes,
        basins=cat_add_lake_old_fdr,
        overwrite=True,
        memory=max_memroy,
    )

    grass.run_command(
        "v.what.rast",
        map=lake_outflow_pourpoints,
        raster=acc,
        column="lmax_acc",
    )
    ### read catchment
    sqlstat = "SELECT cat,lmax_acc FROM %s" % (lake_outflow_pourpoints,)
    lakeinfo = pd.read_sql_query(sqlstat, con)
    lakeinfo = lakeinfo.fillna(-9999)
    lakeinfo = lakeinfo.loc[lakeinfo["cat"] > 0]


    cat_withlake_array = garray.array(mapname=cat_add_lake_old_fdr)
    cat_withlake_array = cat_withlake_array.astype(int)
    fdr_arcgis_array = garray.array(mapname=fdr_arcgis)
    fdr_arcgis_array = fdr_arcgis_array.astype(int)
    str_r_array = garray.array(mapname=str_r)
    sl_lakes_array = garray.array(mapname=sl_lakes)
    sl_lakes_array = sl_lakes_array.astype(int)
    acc_array = garray.array(mapname=acc)
    ncols = int(cat_withlake_array.shape[1])
    nrows = int(cat_withlake_array.shape[0])
    lake_boundary_array = garray.array(mapname=lake_boundary)

    maximumLakegrids = 1000000000
    pec_grid_outlier = 1
    un_modify_fdr_lakeids = []
    outlakeids, chandir, ndir, bd_problem = modify_lakes_flow_direction(
        cat_withlake_array,
        sl_lakes_array,
        acc_array,
        fdr_arcgis_array,
        str_r_array,
        lakeinfo,
        nrows,
        ncols,
        lake_boundary_array,
        pec_grid_outlier,
        maximumLakegrids,
        un_modify_fdr_lakeids,
    )

    temparray = garray.array()

    temparray[:, :] = ndir[:, :]
    temparray.write(mapname=nfdr_arcgis, overwrite=True)
    grass.run_command("r.null", map=nfdr_arcgis, setnull=-9999)

    temparray[:, :] = chandir[:, :]
    temparray.write(mapname="chandir", overwrite=True)
    grass.run_command("r.null", map="chandir", setnull=-9999)

    temparray[:, :] = bd_problem[:, :]
    temparray.write(mapname="bd_problem", overwrite=True)
    grass.run_command("r.null", map="bd_problem", setnull=-9999)

    grass.run_command(
        "r.reclass",
        input=nfdr_arcgis,
        output=nfdr_grass,
        rules=os.path.join(grassdb, "Arcgis2GrassDIR.txt"),
        overwrite=True,
    )
    # cat5
    grass.run_command(
        "r.stream.basins",
        direction=nfdr_grass,
        points=pourpoints_with_lakes,
        basins=cat_add_lake,
        overwrite=True,
        memory=max_memroy,
    )

    grass.run_command("g.copy", rast=(cat_no_lake, cat_use_default_acc), overwrite=True)
    grass.run_command("g.copy", rast=(str_r, "good_seg"), overwrite=True)

    if len(Remove_Str) > 0:
        grass.run_command(
            "r.null", map=cat_use_default_acc, setnull=Remove_Str, overwrite=True
        )
        grass.run_command("r.null", map="good_seg", setnull=Remove_Str, overwrite=True)

        exp = "prob_seg_str = if(isnull(good_seg),%s,null())" % (str_r)
        grass_raster_r_mapcalc(grass, expression=exp)

        exp = "prob_seg_str_w_lid = if(isnull(prob_seg_str),null(),%s)" % (
            sl_connected_lake
        )
        grass_raster_r_mapcalc(grass, expression=exp)

        grass.run_command(
            "g.copy", rast=("prob_seg_str_w_lid", "prob_seg_lake"), overwrite=True
        )

        grass.run_command(
            "r.null",
            map="prob_seg_lake",
            setnull=Lakes_WIth_Multi_Outlet,
            overwrite=True,
        )

        exp = "%s = if(isnull(prob_seg_lake),prob_seg_str_w_lid,null())" % (problem_seg)
        grass_raster_r_mapcalc(grass, expression=exp)

    else:
        exp = "%s = if(isnull(%s),null(),null())" % (problem_seg, "good_seg")
        grass_raster_r_mapcalc(grass, expression=exp)

    grass.run_command(
        "r.to.vect",
        input=problem_seg,
        output=problem_seg,
        type="point",
        overwrite=True,
        flags="v",
    )

    print("Following lake have multi outlet ")
    print(Lakes_WIth_Multi_Outlet)
    print("following str are corrected to make one lake one outlet")
    print(Remove_Str)

    grass.run_command(
        "v.out.ogr",
        input=catchment_pourpoints_outside_lake,
        output=os.path.join(grassdb, catchment_pourpoints_outside_lake + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
        quiet="Ture",
    )
    grass.run_command(
        "v.out.ogr",
        input=pourpoints_with_lakes,
        output=os.path.join(grassdb, pourpoints_with_lakes + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
        quiet="Ture",
    )

    grass.run_command(
        "v.out.ogr",
        input=lake_outflow_pourpoints,
        output=os.path.join(grassdb, lake_outflow_pourpoints + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
        quiet="Ture",
    )
    grass.run_command(
        "v.out.ogr",
        input=lake_inflow_pourpoints,
        output=os.path.join(grassdb, lake_inflow_pourpoints + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
        quiet="Ture",
    )

    PERMANENT.close()
    return lake_outflow_pourpoints



from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import pandas as pd
import numpy as np
import tempfile

def Generate_Grid_Poly_From_NetCDF_QGIS(
    NetCDF_Path="#",
    Output_Folder="#",
    Coor_x_NM="lon",
    Coor_y_NM="lat",
    Is_Rotated_Grid=1,
    R_Coor_x_NM="rlon",
    R_Coor_y_NM="rlat",
    SpatialRef="EPSG:4326",
    x_add=-360,
    y_add=0,
    qgis_prefix_path='#',
):

    """Generate Grid polygon from NetCDF file
    Function that used to generate grid polygon from a NetCDF file
    Parameters
    ----------
    NetCDF_Path                       : string
        It is the path of the NetCDF file
    Output_Folder                     : string
        It is the path to a folder to save output polygon shpfiles
    Coor_x_NM                         : string
        It is the variable name for the x coordinates of grids in
        the NetCDF file
    Coor_y_NM                         : string
        It is the variable name for the y coordinates of grids in
        the NetCDF file
    Is_Rotated_Grid                   : Integer
        1 : indicate the grid in NetCDF file is rotated
        -1: indicate the grid in NetCDF file is not rotated
    R_Coor_x_NM                       : string
        It is the variable name for the y coordinates of rotated
        grids in the NetCDF file
    R_Coor_y_NM                       : string
        It is the variable name for the y coordinates of rotated
        grids in the NetCDF file
    SpatialRef                        : string
        It is the coordinates system used in the NetCDF file
    x_add                             : float
        It is offset value for x coodinate
    y_add                             : float
        It is offset value for y coodinate
    Notes
    -------
    Nc_Grids.shp                      : Point shpfile (output)
       It is point in the center of each netCDF Grids
    Gridncply.shp                     : Polygon shpfile (output)
       It is the polygon for each grid in the NetCDF
    Returns:
    -------
       None
    Examples
    -------
    """

    if not os.path.exists(Output_Folder):
        os.makedirs(Output_Folder)
    tempfolder = os.path.join(
        tempfile.gettempdir(), "basinmaker_" + str(np.random.randint(1, 10000 + 1))
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)


    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)
    from netCDF4 import Dataset

    ncfile = NetCDF_Path
    dsin2 = Dataset(
        ncfile, "r"
    )  # sample structure of in nc file converted from fst

    if Is_Rotated_Grid > 0:
        ncols = len(dsin2.variables[R_Coor_x_NM][:])  ### from 0 to (ncols-1).
        nrows = len(dsin2.variables[R_Coor_y_NM][:])
    else:
        ncols = len(dsin2.variables[Coor_x_NM][:])  ### from 0 to (ncols-1).
        nrows = len(dsin2.variables[Coor_y_NM][:])

    latlonrow = np.full((nrows * ncols, 5), -9999.99999)
    latlonrow = np.full((nrows * ncols, 5), -9999.99999)

    ### Create a point layer, each point will be the nc grids
    cmds = (
        "Point?crs=%s&field=FGID:integer&field=Row:integer&field=Col:integer&field=Gridlon:double&field=Gridlat:double&index=yes"
        % SpatialRef
    )
    Point_Nc_Grid = QgsVectorLayer(cmds, "NC Grid Points", "memory")
    DP_Nc_Point = Point_Nc_Grid.dataProvider()
    Point_Nc_Grid.startEditing()

    ### create polygon layter
    cmds = (
        "Polygon?crs=%s&field=FGID:integer&field=Row:integer&field=Col:integer&field=Gridlon:double&field=Gridlat:double&index=yes"
        % SpatialRef
    )
    Polygon_Nc_Grid = QgsVectorLayer(cmds, "NC Grid polygons", "memory")
    DP_Nc_ply = Polygon_Nc_Grid.dataProvider()
    Polygon_Nc_Grid.startEditing()

    for i in range(0, nrows):
        for j in range(0, ncols):
            k = i * ncols + j

            Point_Fea = QgsFeature()

            if Is_Rotated_Grid < 0:
                latlonrow[k, 0] = k
                latlonrow[k, 1] = i  ### irow
                latlonrow[k, 2] = j  ###col
                latlonrow[k, 3] = dsin2.variables[Coor_x_NM][j] + x_add  ## lon
                latlonrow[k, 4] = dsin2.variables[Coor_y_NM][i]  ## lat

            else:
                latlonrow[k, 0] = k
                latlonrow[k, 1] = i  ### irow
                latlonrow[k, 2] = j  ###col
                latlonrow[k, 3] = dsin2.variables[Coor_x_NM][i, j] + x_add  ## lon
                latlonrow[k, 4] = dsin2.variables[Coor_y_NM][i, j]  ## lat

            #### create point for each grid in Net CDF
            NC_Grid_Point = QgsGeometry.fromPointXY(
                QgsPointXY(latlonrow[k, 3], latlonrow[k, 4])
            )
            Point_Fea.setGeometry(NC_Grid_Point)
            Point_Fea.setAttributes(latlonrow[k, :].tolist())
            DP_Nc_Point.addFeature(Point_Fea)

            ### Create a polygon that the current grid is in the center
            ###  of the polygon

            ### find mid point in row direction
            Polygon_Fea = QgsFeature()
   
            x1 = -1
            x3 = -1 
            x4 = -1
            x2 = -1
            if i != 0 and j != 0:
                if Is_Rotated_Grid > 0:

                    # upper left x1,y1 
                    x1 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j - 1] +
                         dsin2.variables[Coor_x_NM][i-1, j] +
                         dsin2.variables[Coor_x_NM][i-1, j - 1]) / 4 + x_add
                         
                    y1 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j - 1] +
                         dsin2.variables[Coor_y_NM][i-1, j] +
                         dsin2.variables[Coor_y_NM][i-1, j - 1]) / 4
                         
                         
            if i!= 0 and j != ncols -1:
                if Is_Rotated_Grid > 0:

                    x2 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i-1, j] +
                         dsin2.variables[Coor_x_NM][i-1, j + 1] +
                         dsin2.variables[Coor_x_NM][i, j + 1]) / 4 + x_add              
                    y2 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i-1, j] +
                         dsin2.variables[Coor_y_NM][i-1, j + 1] +
                         dsin2.variables[Coor_y_NM][i, j + 1]) / 4 
            
            if i != nrows - 1 and j != ncols -1:
                if Is_Rotated_Grid > 0: 
                    # bot right x3,y3 
                    x3 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j +1] +
                         dsin2.variables[Coor_x_NM][i+1, j + 1] +
                         dsin2.variables[Coor_x_NM][i + 1, j]) / 4 + x_add

                    y3 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j +1] +
                         dsin2.variables[Coor_y_NM][i+1, j + 1] +
                         dsin2.variables[Coor_y_NM][i + 1, j]) / 4
                         
            if i != nrows - 1 and j != 0 :                            
                if Is_Rotated_Grid > 0:
                    x4 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j - 1] +
                         dsin2.variables[Coor_x_NM][i+1, j - 1] +
                         dsin2.variables[Coor_x_NM][i + 1, j]) / 4 + x_add

                    y4 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j - 1] +
                         dsin2.variables[Coor_y_NM][i+1, j - 1] +
                         dsin2.variables[Coor_y_NM][i + 1, j]) / 4
            
            # p4 and p1 and p2    
            if i == 0 and j == 0:
                if Is_Rotated_Grid > 0:
                    # x1 = dsin2.variables[Coor_x_NM][i, j]+ x_add  
                    # y1 = dsin2.variables[Coor_y_NM][i, j]
                    
                    x2_t = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j + 1]) / 2 + x_add              
                    y2_t = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j + 1]) / 2 

                    x2 = x2_t + (x2_t - x3)
                    y2 = y2_t + (y2_t - y3)
                    
                    x4 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i + 1, j]) / 2 + x_add

                    y4 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i + 1, j]) / 2

            if i == 0 and j != 0 and j != ncols -1 :
                if Is_Rotated_Grid > 0:
                    x1 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j - 1]) / 2 + x_add
                         
                    y1 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j - 1]) / 2
                    
                    x2 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j + 1]) / 2 + x_add              
                    y2 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j + 1]) / 2 
                         
            if i == 0 and j ==  ncols -1:
                if Is_Rotated_Grid > 0:
                    x1 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j - 1]) / 2 + x_add
                         
                    y1 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j - 1]) / 2

                    x2 = dsin2.variables[Coor_x_NM][i, j]+ x_add  
                    y2 = dsin2.variables[Coor_y_NM][i, j]                    

                    x3 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i + 1, j]) / 2 + x_add

                    y3 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i + 1, j]) / 2                                                                        
            
            if i == nrows - 1 and j != 0 and j != ncols -1:
                if Is_Rotated_Grid > 0:
                    x3 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j +1]) / 2 + x_add

                    y3 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j +1] ) / 2   
                                                                                        
                    x4 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j - 1]) / 2 + x_add

                    y4 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j - 1]) / 2

            if i == nrows - 1 and j == 0:
                if Is_Rotated_Grid > 0:
            
                    x4 = dsin2.variables[Coor_x_NM][i, j] + x_add
            
                    y4 = dsin2.variables[Coor_y_NM][i, j]
            
                    x1 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i-1, j]) / 2 + x_add
                    y1 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i-1, j]) / 2
            
                    x3 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j +1]) / 2 + x_add
            
                    y3 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j +1]) / 2
                                                                                                                                                              

                         
            if i == nrows - 1 and j == ncols -1:
                if Is_Rotated_Grid > 0:
            
                    x3 = dsin2.variables[Coor_x_NM][i, j] + x_add
            
                    y3 = dsin2.variables[Coor_y_NM][i, j]
            
                    x2 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i-1, j]) / 2 + x_add              
                    y2 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i-1, j]) / 2
            
                    x4 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i, j - 1]) / 2 + x_add
            
                    y4 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i, j - 1]) / 2


            if i != nrows - 1 and i !=0 and j == 0:
                if Is_Rotated_Grid > 0:

                    x1 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i-1, j]) / 2 + x_add
                         
                    y1 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i-1, j]) / 2
                                     
                         
                    x4 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i + 1, j]) / 2 + x_add

                    y4 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i + 1, j]) / 2
                         

            if i != nrows - 1 and i !=0 and j == ncols -1:
                if Is_Rotated_Grid > 0:

                    x2 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i-1, j] ) / 2 + x_add              
                    y2 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i-1, j]) / 2
                                                                                                                                             
                    x3 = (dsin2.variables[Coor_x_NM][i, j] +
                         dsin2.variables[Coor_x_NM][i + 1, j]) / 2 + x_add

                    y3 = (dsin2.variables[Coor_y_NM][i, j] +
                         dsin2.variables[Coor_y_NM][i + 1, j]) / 2                    

            if x1 == -1 or x3 == -1 or x4 == -1 or x2 == -1:
                continue
            Point_1 = QgsPointXY(x1, y1)  ## lower left
            Point_2 = QgsPointXY(x2, y2)
            Point_3 = QgsPointXY(x3, y3)
            Point_4 = QgsPointXY(x4, y4)
            # if i == nrows - 1:
            #     print("#########################################")
            #     print(x1, y1)
            #     print(x2, y2)
            #     print(x3, y3)
            #     print(x4, y4)
            #     print(dsin2.variables[Coor_x_NM][i, j] + x_add,dsin2.variables[Coor_y_NM][i, j])
            #     print("#########################################")
            
            gPolygon = QgsGeometry.fromPolygonXY(
                [[Point_1, Point_2, Point_3, Point_4]]
            )
            Polygon_Fea.setGeometry(gPolygon)
            Polygon_Fea.setAttributes(latlonrow[k, :].tolist())
            DP_Nc_ply.addFeature(Polygon_Fea)

    Point_Nc_Grid.commitChanges()
    Point_Nc_Grid.updateExtents()

    Polygon_Nc_Grid.commitChanges()
    Polygon_Nc_Grid.updateExtents()

    pdlatlonrow = pd.DataFrame(
        latlonrow, columns=["FGID", "Row", "Col", "Gridlon", "Gridlat"]
    )
    pdlatlonrow.to_csv(
        os.path.join(Output_Folder, "Gridcorr.csv"), sep=",", index=False
    )

    QgsVectorFileWriter.writeAsVectorFormat(
        layer=Point_Nc_Grid,
        fileName=os.path.join(Output_Folder, "Nc_Grids.shp"),
        fileEncoding="UTF-8",
        destCRS=QgsCoordinateReferenceSystem(SpatialRef),
        driverName="ESRI Shapefile",
    )
    QgsVectorFileWriter.writeAsVectorFormat(
        layer=Polygon_Nc_Grid,
        fileName=os.path.join(Output_Folder, "Gridncply.shp"),
        fileEncoding="UTF-8",
        destCRS=QgsCoordinateReferenceSystem(SpatialRef),
        driverName="ESRI Shapefile",
    )

def Area_Weighted_Mapping_Between_Two_Polygons_QGIS(
    Target_Ply_Path="#",
    Mapping_Ply_Path="#",
    Col_NM="HRU_ID",
    Output_Folder="#",
    qgis_prefix_path = '#',
):

    """Generate Grid polygon from NetCDF file
    Function that used to generate grid polygon from a NetCDF file
    Parameters
    ----------
    Target_Ply_Path                       : string
        It is the path of one inputs HRU polygon file
    Mapping_Ply_Path                      : string
        It is the path of one inputs grid polygon file
    Output_Folder                     : string
        It is the path to a folder to save output polygon shpfiles
    Notes
    -------
    Overlay_Polygons.shp                 : Polygon shpfile (output)
       It is overlay of two input polygons shpfiles
    GriddedForcings2.txt                 : Text file (output)
       It is the polygon area weighted of each polygon in Mapping_Ply_Path
       to each polygon in Target_Ply_Path
    Returns:
    -------
       None
    Examples
    -------
    """
    
    if not os.path.exists(Output_Folder):
        os.makedirs(Output_Folder)
        
    tempfolder = os.path.join(
        tempfile.gettempdir(), "basinmaker_" + str(np.random.randint(1, 10000 + 1))
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    Path_finalcat_hru_temp = os.path.join(
        tempfolder,
        str(np.random.randint(1, 10000 + 1)) + "finalcat_freferen.shp",
    )
    Path_finalcat_hru_temp2 = os.path.join(
        tempfolder,
        str(np.random.randint(1, 10000 + 1)) + "finalcat_freferen2.shp",
    )
    Path_finalcat_hru_temp_dissolve = os.path.join(
        tempfolder,
        str(np.random.randint(1, 10000 + 1)) + "finalcat_freferen_dissolve.shp",
    )
    Path_finalcat_hru_temp_dissolve_area = os.path.join(
        Output_Folder, "Overlay_Polygons.shp"
    )
    
    Target_Ply_mem = qgis_vector_fix_geometries(
        processing, context, INPUT=Target_Ply_Path, OUTPUT="memory:"
    )["OUTPUT"]

    Mapping_Ply_mem = qgis_vector_fix_geometries(
        processing, context, INPUT=Mapping_Ply_Path, OUTPUT="memory:"
    )["OUTPUT"]
            
    ### create overlay betweeo two polygon and calcuate area of
    ### each new polygon in the overlay
    qgis_vector_union_two_layers(
        processing=processing,
        context=context,
        INPUT=Target_Ply_mem,
        OVERLAY=Mapping_Ply_mem,
        OVERLAY_FIELDS_PREFIX="Map_",
        OUTPUT=Path_finalcat_hru_temp,
    )["OUTPUT"]

    finalcat_hru_temp2 = processing.run(
        "native:extractbyattribute",
        {
            "INPUT": Path_finalcat_hru_temp,
            "FIELD": "HRU_ID",
            "OPERATOR": 2,
            "VALUE": "0",
            "OUTPUT": "memory:",
        },
    )["OUTPUT"]
    processing.run(
        "native:extractbyattribute",
        {
            "INPUT": finalcat_hru_temp2,
            "FIELD": "Map_FGID",
            "OPERATOR": 2,
            "VALUE": "0",
            "OUTPUT": Path_finalcat_hru_temp2,
        },
    )
        
    processing.run(
        "native:dissolve",
        {
            "INPUT": Path_finalcat_hru_temp2,
            "FIELD": ["HRU_ID", "Map_FGID"],
            "OUTPUT": Path_finalcat_hru_temp_dissolve,
        },
        context=context,
    )
    qgis_vector_field_calculator(
        processing=processing,
        context=context,
        FORMULA="area(transform($geometry, 'EPSG:3161','EPSG:3161'))",
        FIELD_NAME="s_area",
        INPUT=Path_finalcat_hru_temp_dissolve,
        OUTPUT=Path_finalcat_hru_temp_dissolve_area,
        FIELD_PRECISION=3,
    )["OUTPUT"]

    ### calculate the area weight of the mapping polygon to target polygon

    dbf1 = Dbf5(Mapping_Ply_Path[:-3] + "dbf")
    Forcinfo = dbf1.to_dataframe()
    Avafgid = Forcinfo["FGID"].values

    dbf2 = Dbf5(Path_finalcat_hru_temp_dissolve_area[:-3] + "dbf")
    Mapforcing = dbf2.to_dataframe()
    Mapforcing = Mapforcing.loc[Mapforcing[Col_NM] > 0]  ### remove
    Mapforcing = Mapforcing.loc[Mapforcing["Map_FGID"] > 0]  ### remove
    Mapforcing = Mapforcing.loc[Mapforcing["s_area"] > 0.000001]  ### remove
    
    grid_weight_string = create_grid_weight_main(Mapforcing,Forcinfo)
    grid_weight_file_path = os.path.join(Output_Folder, "GriddedForcings2.txt")
    WriteStringToFile(grid_weight_string, grid_weight_file_path, "w")



from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import pandas as pd
import numpy as np
import shutil
import tempfile


def simplify_routing_structure_by_drainage_area_qgis(
    Routing_Product_Folder= '#',
    Area_Min=-1,
    OutputFolder="#",
    qgis_prefix_path="#",
):
    """Simplify the routing product by drainage area

    Function that used to simplify the routing product by
    using user provided minimum subbasin drainage area.
    The input catchment polygons are routing product before
    merging for lakes. It is provided with routing product.
    The result is the simplified catchment polygons. But
    result from this fuction still not merging catchment
    covering by the same lake. Thus, The result generated
    from this tools need further processed by
    Define_Final_Catchment, or can be further processed by
    SelectLakes

    Parameters
    ----------

    Path_final_riv_ply             : string
        Path to the catchment polygon which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.
    Path_final_riv                 : string
        Path to the river polyline which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.
    Path_Con_Lake_ply              : string
        Path to a connected lake polygon. Connected lakes are lakes that
        are connected by Path_final_riv.
    Path_NonCon_Lake_ply           : string
        Path to a non connected lake polygon. Connected lakes are lakes
        that are not connected by Path_final_riv.
    Area_Min                       : float
        The minimum drainage area of each catchment in km2
    OutputFolder                   : string
        Folder name that stores generated simplified routing product

    Notes
    -------
    This function has no return values, instead will generate following
    files. The output tpye will be the same as inputs, but the routing
    network will be simplified by increase subbasin size, reduce
    number of subbasins and number of river segments.

    os.path.join(OutputFolder,os.path.basename(Path_final_riv_ply))
    os.path.join(OutputFolder,os.path.basename(Path_final_riv))
    os.path.join(OutputFolder,os.path.basename(Path_Con_Lake_ply))
    os.path.join(OutputFolder,os.path.basename(Path_NonCon_Lake_ply))

    Returns:
    -------
    None

    Examples
    -------

    """
    #### generate river catchments based on minmum area.
    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    sub_colnm = "SubId"
    down_colnm = "DowSubId"
    DA_colnm = "DrainArea"
    SegID_colnm = "Seg_ID"

    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)

    tempfolder = os.path.join(
        tempfile.gettempdir(), "basinmaker_inda_" + str(np.random.randint(1, 10000 + 1))
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)

    Path_Catchment_Polygon="#"
    Path_River_Polyline="#"
    Path_Con_Lake_ply="#"
    Path_NonCon_Lake_ply="#"
    Path_obs_gauge_point="#"
    Path_final_cat_ply="#"
    Path_final_cat_riv="#"

    ##define input files from routing prodcut 
    for file in os.listdir(Routing_Product_Folder):
        if file.endswith(".shp"):
            if 'catchment_without_merging_lakes' in file:
                Path_Catchment_Polygon = os.path.join(Routing_Product_Folder, file)
            if 'river_without_merging_lakes' in file:
                Path_River_Polyline = os.path.join(Routing_Product_Folder, file)
            if 'sl_connected_lake' in file:
                Path_Con_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'sl_non_connected_lake' in file:
                Path_NonCon_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'obs_gauges' in file:
                Path_obs_gauge_point = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info' in file:
                Path_final_cat_ply = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info_riv' in file:
                Path_final_cat_riv = os.path.join(Routing_Product_Folder, file)                

    if Path_Catchment_Polygon == '#' or  Path_River_Polyline =='#':
        print("Invalid routing product folder ")

    Path_final_riv_ply = Path_Catchment_Polygon
    Path_final_riv = Path_River_Polyline

    ## copy obs_gauges to output folder 
    for file in os.listdir(Routing_Product_Folder):
        if 'obs_gauges' in file:
            shutil.copy(os.path.join(Routing_Product_Folder, file), os.path.join(OutputFolder, file))



    # overall procedure,
    # 1. first get product attribute table
    # 2. determine which features needs to be merged together to increage
    #    drainage area of the sub basin, for example sub a, b c needs to be merged
    # 3. in the attribute table, change sub a b c 's content to a, assuming sub b and c drainge to a.
    # 4. copy modified attribute table to shafiles
    # 5. dissolve based on subid amd finished
    # overall procedure,

    # read attribute table, and
    Path_final_rviply = Path_final_riv_ply
    Path_final_riv = Path_final_riv
    Path_Conl_ply = Path_Con_Lake_ply
    Path_Non_ConL_ply = Path_NonCon_Lake_ply

    finalriv_info = Dbf_To_Dataframe(Path_final_rviply).drop_duplicates(
        sub_colnm, keep="first"
    )

    if Path_Conl_ply != '#':
        Conn_Lakes_ply = Dbf_To_Dataframe(Path_Conl_ply).drop_duplicates(
            "Hylak_id", keep="first"
        )
    else: 
        Conn_Lakes_ply = pd.DataFrame(np.full((10,1),-9999),columns=["Hylak_id"])
    # change attribute table
    (
        mapoldnew_info,
        Selected_riv_ids,
        Connected_Lake_Mainriv,
        Old_Non_Connect_LakeIds,
        Conn_To_NonConlakeids,
    ) = Change_Attribute_Values_For_Catchments_Need_To_Be_Merged_By_Increase_DA(
        finalriv_info, Conn_Lakes_ply, Area_Min
    )

    # remove river polyline
    Path_Temp_final_rviply = os.path.join(tempfolder, "temp1_finalriv_ply.shp")
    Path_Temp_final_rvi = os.path.join(tempfolder, "temp1_finalriv.shp")
    Selectfeatureattributes(
        processing,
        Input=Path_final_riv,
        Output=Path_Temp_final_rvi,
        Attri_NM="SubId",
        Values=Selected_riv_ids,
    )
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=Path_final_rviply,
        FIELD=["SubId"],
        OUTPUT=Path_Temp_final_rviply,
    )

    # update topology of new attribute table
    UpdateTopology(mapoldnew_info)
    mapoldnew_info = update_non_connected_catchment_info(mapoldnew_info)

    # create output folder
    outputfolder_subid = OutputFolder
    if not os.path.exists(outputfolder_subid):
        os.makedirs(outputfolder_subid)
        
    all_subids = finalriv_info['SubId'].values
    
    copy_data_and_dissolve(all_subids,tempfolder,processing,Path_Temp_final_rviply,Path_Temp_final_rvi,
        mapoldnew_info,COLUMN_NAMES_CONSTANT_CLEAN,OutputFolder,Path_Catchment_Polygon,context,
        Path_final_rviply,Path_final_riv)
        
        
    # # copy new attribute table to subbasin polyline and polygon
    # Copy_Pddataframe_to_shpfile(
    #     Path_Temp_final_rviply,
    #     mapoldnew_info,
    #     link_col_nm_shp="SubId",
    #     link_col_nm_df="Old_SubId",
    #     UpdateColNM=["#"],
    # )
    # Copy_Pddataframe_to_shpfile(
    #     Path_Temp_final_rvi,
    #     mapoldnew_info,
    #     link_col_nm_shp="SubId",
    #     link_col_nm_df="Old_SubId",
    #     UpdateColNM=["#"],
    # )



    # export lake polygons
    if Path_Conl_ply != '#' and len(Connected_Lake_Mainriv) > 0:
        Selectfeatureattributes(
            processing,
            Input=Path_Conl_ply,
            Output=os.path.join(outputfolder_subid, os.path.basename(Path_Conl_ply)),
            Attri_NM="Hylak_id",
            Values=Connected_Lake_Mainriv,
        )
    if Path_Non_ConL_ply != '#' and len(Old_Non_Connect_LakeIds) >0:
        Selectfeatureattributes(
            processing,
            Input=Path_Non_ConL_ply,
            Output=os.path.join(outputfolder_subid, os.path.basename(Path_Non_ConL_ply)),
            Attri_NM="Hylak_id",
            Values=Old_Non_Connect_LakeIds,
        )

    # Copy connected lakes that are transfered into non-connected
    # lake to non connected lake polygon

    if len(Conn_To_NonConlakeids) > 0 and Path_Conl_ply !='#':
        if Path_Non_ConL_ply != '#':
            Copyfeature_to_another_shp_by_attribute(
                Source_shp=Path_Conl_ply,
                Target_shp=os.path.join(
                    outputfolder_subid, os.path.basename(Path_Non_ConL_ply)
                ),
                Col_NM="Hylak_id",
                Values=Conn_To_NonConlakeids,
                Attributes=Conn_Lakes_ply,
            )
        else:
            if len(os.path.basename(Path_Conl_ply).split('_')) == 4:
                outlake_name = 'sl_non_connected_lake_' + os.path.basename(Path_Conl_ply).split('_')[3]
            else:
                outlake_name = 'sl_non_connected_lake.shp'
            Selectfeatureattributes(
                processing,
                Input=Path_Conl_ply,
                Output=os.path.join(outputfolder_subid, outlake_name),
                Attri_NM="Hylak_id",
                Values=Conn_To_NonConlakeids,
            )
                    
    # # dissolve subbasin polygon based on new subbasin id
    # Path_out_final_rviply = os.path.join(
    #     outputfolder_subid, os.path.basename(Path_final_riv_ply)
    # )
    # Path_out_final_rvi = os.path.join(
    #     outputfolder_subid, os.path.basename(Path_final_riv)
    # )
    # 
    # qgis_vector_dissolve(
    #     processing,
    #     context,
    #     INPUT=Path_Temp_final_rviply,
    #     FIELD=["SubId"],
    #     OUTPUT=Path_out_final_rviply,
    # )
    # qgis_vector_dissolve(
    #     processing,
    #     context,
    #     INPUT=Path_Temp_final_rvi,
    #     FIELD=["SubId"],
    #     OUTPUT=Path_out_final_rvi,
    # )
    # 
    # # clean attribute table and done
    # Clean_Attribute_Name(Path_out_final_rviply, COLUMN_NAMES_CONSTANT_CLEAN)
    # Clean_Attribute_Name(Path_out_final_rvi, COLUMN_NAMES_CONSTANT_CLEAN)

    Qgs.exit()



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import os


def obtain_polygon_boundary(grassdb, qgis_prefix_path, ply_path, output):
    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    # obtain lake boundary lines
    qgis_vector_polygon_stro_lines(
        processing,
        context,
        INPUT=ply_path,
        OUTPUT=output,
    )


def reproject_clip_vectors_by_polygon(
    grassdb, grass_location, qgis_prefix_path, mask, path_polygon, ply_name
):

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    crs_id = qgis_vector_return_crs_id(
        processing, context, mask, Input_Is_Feature_In_Mem=False
    )

    qgis_vector_reproject_layers(
        processing,
        context,
        INPUT=path_polygon,
        TARGET_CRS=crs_id,
        OUTPUT=os.path.join(grassdb, ply_name + "_project.shp"),
    )
    # lake polygon sometime has error in geometry
    try:
        qgis_vector_ectract_by_location(
            processing,
            context,
            INPUT=os.path.join(grassdb, ply_name + "_project.shp"),
            INTERSECT=mask,
            OUTPUT=os.path.join(grassdb, ply_name + ".shp"),
        )
        if ply_name == 'kc_zone':
            processing.run("native:clip", {
                           'INPUT':os.path.join(grassdb, ply_name + "_project.shp"),
                           'OVERLAY':mask,
                           'OUTPUT':os.path.join(grassdb, ply_name + "_clip.shp")})

    except:
        print("Need fix lake boundary geometry to speed up")
        qgis_vector_fix_geometries(
            processing,
            context,
            INPUT=os.path.join(grassdb, ply_name + "_project.shp"),
            OUTPUT=os.path.join(grassdb, ply_name + "_fixgeo.shp"),
        )
        qgis_vector_fix_geometries(
            processing,
            context,
            INPUT=mask,
            OUTPUT=os.path.join(grassdb, "mask_fixgeo.shp"),
        )
        
        qgis_vector_ectract_by_location(
            processing,
            context,
            INPUT=os.path.join(grassdb, ply_name + "_fixgeo.shp"),
            INTERSECT=os.path.join(grassdb, "mask_fixgeo.shp"),
            OUTPUT=os.path.join(grassdb, ply_name + ".shp"),
        )

        if ply_name == 'kc_zone':
            processing.run("native:clip", {
                           'INPUT':os.path.join(grassdb, ply_name + "_fixgeo.shp"),
                           'OVERLAY':os.path.join(grassdb, "mask_fixgeo.shp"),
                           'OUTPUT':os.path.join(grassdb, ply_name + "_clip.shp")})
                       
    Qgs.exit()
